"""Probability of persistent contrail coverage (PCC)."""

from __future__ import annotations

import dataclasses
from typing import Any

import numpy as np
import xarray as xr

from pycontrails.core.fuel import Fuel, JetA
from pycontrails.core.met import MetDataArray, MetDataset
from pycontrails.core.met_var import AirTemperature, SpecificHumidity
from pycontrails.core.models import Model, ModelParams
from pycontrails.datalib.ecmwf.variables import SpecificCloudIceWaterContent
from pycontrails.models import sac
from pycontrails.models.humidity_scaling import HumidityScaling
from pycontrails.physics import thermo


@dataclasses.dataclass
class PCCParams(ModelParams):
    """PCC Model Parameters."""

    #: Cloud model
    #: Options include "Smith1990", "Sundqvist1989", "Slingo1980"
    cloud_model: str = "Smith1990"

    #: Critical RH Factor for the model to cirrus clouds
    rh_crit_factor: float = 0.7

    #: Fuel type
    fuel: Fuel = dataclasses.field(default_factory=JetA)

    #: Engine efficiency
    engine_efficiency: float = 0.35

    #: Humidity scaling
    humidity_scaling: HumidityScaling | None = None


class PCC(Model):
    r"""Potential Contrail Coverage Algorithm.

    Determines the potential of ambient atmosphere to allow contrail formation at grid points.

    Parameters
    ----------
    met : MetDataset
        Dataset containing  :attr:`met_variables` variables.
    surface : MetDataset
        Surface level dataset containing "air_pressure".
    params : dict[str, Any], optional
        Override PCC model parameters with dictionary.
        See :class:`PCCParams` for model parameters.
    **params_kwargs
        Override PCC model parameters with keyword arguments.
        See :class:`PCCParams` for model parameters.

    Notes
    -----
    Based on Ponater et al. (2002)
    """

    __slots__ = ("_cloud_model", "surface")
    name = "pcc"
    long_name = "Potential contrail coverage"
    met_variables = (AirTemperature, SpecificHumidity, SpecificCloudIceWaterContent)
    _cloud_model: Any
    source: MetDataset
    surface: MetDataset
    default_params = PCCParams

    def __init__(
        self,
        met: MetDataset,
        surface: MetDataset,
        params: dict[str, Any] = {},
        **params_kwargs: Any,
    ) -> None:
        super().__init__(met, params=params, **params_kwargs)

        # set cloud model by method
        self._cloud_model = getattr(self, self.params["cloud_model"], None)
        if self._cloud_model is None:
            raise ValueError(
                "Cloud model must be one of 'Smith1990', 'Sundqvist1989', 'Slingo1980'"
            )

        # make sure surface level dataset is valid
        if not isinstance(surface, MetDataset):
            raise ValueError("Surface air pressure required as input parameter `surface`")

        self.surface = surface.copy()
        self.surface.ensure_vars(["surface_air_pressure"])

    def eval(self, source: MetDataset | None = None, **params: Any) -> MetDataArray:
        """Evaluate PCC model.

        Currently only implemented to work on the :attr:`met` data input.

        Parameters
        ----------
        source : MetDataset | None, optional
           Input MetDataset.
            If None, evaluates at the :attr:`met` grid points.
        **params : Any
            Overwrite model parameters before eval


        Returns
        -------
        MetDataArray
            PCC model output
        """
        self.update_params(params)
        self.set_source(source)

        # apply humidity scaling
        scale_humidity = (self.params["humidity_scaling"] is not None) and (
            "specific_humidity" not in self.source
        )
        self.set_source_met()

        if scale_humidity:
            self.params["humidity_scaling"].eval(self.source, copy_source=False)

        return MetDataArray(self.b_contr(), name="pcc")

    def b_contr(self) -> xr.DataArray:
        """Calculate critical relative humidity threshold of contrail formation.

        Returns
        -------
        xr.DataArray
            Critical relative humidity of contrail formation, [:math:`[0 - 1]`]

        Notes
        -----
        Instead of using a prescribed threshold relative humidity for ``rh_crit_old``
        the threshold relative humidity now change with pressure.

        This equation is described in Roeckner et al. 1996, Eq.57
        THE ATMOSPHERIC GENERAL CIRCULATION MODEL ECHAM-4: MODEL DESCRIPTION AND
        SIMULATION OF PRESENT-DAY CLIMATE
        """
        sp = self.surface["surface_air_pressure"].data.loc[dict(level=-1)]  # surface air pressure

        def _apply_b_contr_plev(_ds: xr.Dataset) -> xr.Dataset:
            # cannot instantiate SAC object unless all 4 x, y, z, t dimensions are present
            # convert "level" coordinate to dimension
            _ds = _ds.expand_dims(dim="level")

            p = _ds["air_pressure"]

            G = sac.slope_mixing_line(
                _ds["specific_humidity"],
                _ds["air_pressure"],
                self.params["engine_efficiency"],
                self.params["fuel"].ei_h2o,
                self.params["fuel"].q_fuel,
            )
            T_sat_liquid = sac.T_sat_liquid(G)
            rh_crit_sac = sac.rh_critical_sac(_ds["air_temperature"], T_sat_liquid, G)

            # rh_crit_old = np.ones(p.shape) * self.rh_crit_factor
            rh_crit_old = 0.99 + (0.6 - 0.99) * np.exp(1 - (sp / p) ** 4)
            rh_crit_new = rh_crit_sac * rh_crit_old

            b_crit = self._cloud_model(
                _ds["air_temperature"],
                _ds["air_pressure"],
                _ds["specific_cloud_ice_water_content"],
                _ds["specific_humidity"],
                rh_crit_old,
                rh_crit_old,
            )
            b_crit_contr = self._cloud_model(
                _ds["air_temperature"],
                _ds["air_pressure"],
                _ds["specific_cloud_ice_water_content"],
                _ds["specific_humidity"],
                rh_crit_old,
                rh_crit_new,
            )
            b_crit_potential = b_crit_contr - b_crit

            b_crit_potential = xr.where((b_crit_potential > 1), 1, b_crit_potential)
            b_crit_potential = xr.where((b_crit_potential < 0), 0, b_crit_potential)

            # issue recombining groups arises if "level" is in dims
            # convert "level" dimension to coordinate
            b_crit_potential = b_crit_potential.squeeze("level")
            return b_crit_potential

        # apply calculation per pressure level
        return (
            self.source.data.groupby("level")  # type: ignore
            .map(_apply_b_contr_plev)
            .transpose(*self.source.dim_order)
        )

    def Smith1990(
        self,
        T: xr.DataArray,
        p: xr.DataArray,
        iwc: xr.DataArray,
        q: xr.DataArray,
        rh_crit_old: xr.DataArray,
        rh_crit_new: xr.DataArray,
    ) -> xr.DataArray:
        r"""Apply Smith Scheme described in Rap et al. (2009).

        Parameterization of contrails in the UK Met OfficeClimate Model;

        Parameters
        ----------
        T : :class:`xarray:DataArray`
            Air Temperature, [:math:`K`]
        p : :class:`xarray:DataArray`
            Air Pressure, [:math:`Pa`]
        iwc : :class:`xarray:DataArray`
            Cloud ice water content, [:math:`kg \ kg^{-1}`]
        q : :class:`xarray:DataArray`
            Specific humidity
        rh_crit_old : :class:`xarray:DataArray`
            Critical relative humidity, [:math:`[0 - 1]`]
        rh_crit_new : :class:`xarray:DataArray`
            Critical relative humidity, [:math:`[0 - 1]`]

        Returns
        -------
        :class:`xarray:DataArray`
            Probability of cirrus formation, [:math:`[0 - 1]`]
        """
        r = thermo.rh(q, T, p)
        q_sw = thermo.q_sat(T, p)
        b_crit = (1 - rh_crit_old) * q_sw
        Q_n = iwc / b_crit - (1 - r) / (1 - rh_crit_new)

        b_cirrus = xr.DataArray(np.zeros(Q_n.shape), coords=Q_n.coords)
        b_cirrus = xr.where((Q_n <= -1), 0, b_cirrus)
        b_cirrus = xr.where(
            (Q_n > -1) & (Q_n <= 0), 0.5 * (1 + Q_n.where((Q_n > -1) & (Q_n <= 0))) ** 2, b_cirrus
        )
        b_cirrus = xr.where(
            (Q_n > 0) & (Q_n <= 1), 1 - 0.5 * (1 - Q_n.where((Q_n > 0) & (Q_n <= 1))) ** 2, b_cirrus
        )
        return xr.where((Q_n > 1), 1, b_cirrus)

    def Slingo1980(
        self,
        T: xr.DataArray,
        p: xr.DataArray,
        iwc: xr.DataArray,
        q: xr.DataArray,
        rh_crit_old: xr.DataArray,
        rh_crit_new: xr.DataArray,
    ) -> xr.DataArray:
        r"""Apply Slingo scheme described in Wood and Field, 1999.

        Relationships between Total Water, Condensed Water, and Cloud Fraction in
        Stratiform Clouds Examined Using Aircraft Data

        Parameters
        ----------
        T : :class:`xarray:DataArray`
            Air Temperature, [:math:`K`]
        p : :class:`xarray:DataArray`
            Air Pressure, [:math:`Pa`]
        iwc : :class:`xarray:DataArray`
            Cloud ice water content, [:math:`kg \ kg^{-1}`]
        q : :class:`xarray:DataArray`
            Specific humidity
        rh_crit_old : :class:`xarray:DataArray`
            Critical relative humidity, [:math:`[0 - 1]`]
        rh_crit_new : :class:`xarray:DataArray`
            Critical relative humidity, [:math:`[0 - 1]`]

        Returns
        -------
        :class:`xarray:DataArray`
            Probability of cirrus formation, [:math:`[0 - 1]`]
        """
        r = thermo.rh(q, T, p)
        b_cirrus = ((r - rh_crit_new) / (1 - rh_crit_new)) ** 2
        b_cirrus = xr.where(r < rh_crit_new, 0, b_cirrus)
        return xr.where(r >= 1.0, 1.0, b_cirrus)

    def Sundqvist1989(
        self,
        T: xr.DataArray,
        p: xr.DataArray,
        iwc: xr.DataArray,
        q: xr.DataArray,
        rh_crit_old: xr.DataArray,
        rh_crit_new: xr.DataArray,
    ) -> xr.DataArray:
        r"""Apply Sundqvist scheme described in Ponater et al. (2002).

        Contrails in a comprehensive global climate model: Parameterization and radiative
        forcing results

        Parameters
        ----------
        T : :class:`xarray:DataArray`
            Air Temperature, [:math:`K`]
        p : :class:`xarray:DataArray`
            Air Pressure, [:math:`Pa`]
        iwc : :class:`xarray:DataArray`
            Cloud ice water content, [:math:`kg \ kg^{-1}`]
        q : :class:`xarray:DataArray`
            Specific humidity
        rh_crit_old : :class:`xarray:DataArray`
            Critical relative humidity, [:math:`[0 - 1]`]
        rh_crit_new : :class:`xarray:DataArray`
            Critical relative humidity, [:math:`[0 - 1]`]

        Returns
        -------
        :class:`xarray:DataArray`
            Probability of cirrus formation, [:math:`[0 - 1]`]
        """
        r = thermo.rh(q, T, p)

        # clipping ratio at 1 to prevent np.sqrt from taking negative arguments
        ratio = (r - rh_crit_new) / (1 - rh_crit_new)
        ratio = xr.where(ratio > 1, 1, ratio)

        b_cirrus = 1 - (1 - ratio) ** 0.5
        return xr.where(r < rh_crit_new, 0, b_cirrus)



"""Test tau_cirrus module."""

from __future__ import annotations

from typing import Callable

import numpy as np
import pytest
import xarray as xr

from pycontrails import MetDataset
from pycontrails.models.tau_cirrus import tau_cirrus
from tests._deprecated import tau_cirrus_original


@pytest.mark.parametrize("formula", [tau_cirrus, tau_cirrus_original])
def test_shape_and_coords(met_cocip1: MetDataset, formula: Callable) -> None:
    """Ensure shape and coords are consistent."""
    met_cocip1.data["air_pressure"] = met_cocip1.data["air_pressure"].astype(np.float32)

    da = formula(met_cocip1)
    assert isinstance(da, xr.DataArray)

    # As long as the variables and air_pressure are float32, the output should also be float32
    assert da.dtype == np.float32

    assert da.sizes == met_cocip1.data.sizes
    assert list(da.dims) == met_cocip1.dim_order

    xr.testing.assert_equal(da.level, met_cocip1.data["level"])
    xr.testing.assert_equal(da.time, met_cocip1.data["time"])
    xr.testing.assert_equal(da.longitude, met_cocip1.data["longitude"])
    xr.testing.assert_equal(da.latitude, met_cocip1.data["latitude"])

    assert da.name == "tau_cirrus"


@pytest.mark.parametrize("formula", [tau_cirrus, tau_cirrus_original])
def test_nonnegative_values(met_cocip1: MetDataset, formula: Callable) -> None:
    """Ensure tau cirrus values are nonnegative."""
    da = formula(met_cocip1)
    assert (da >= 0).all()


def test_implementations_close(met_cocip1: MetDataset) -> None:
    """Ensure current implementation is close to original implementation."""
    met_cocip1.data["air_pressure"] = met_cocip1.data["air_pressure"].astype(np.float32)

    # Both are nonzero in same places
    da1 = tau_cirrus(met_cocip1)
    da2 = tau_cirrus_original(met_cocip1)
    xr.testing.assert_equal(da1 > 0, da2 > 0)

    # New implementation is larger
    assert (da1 >= da2).all()

    # But not twice as large
    assert (da1 <= 2 * da2).all()

    # Pin the mean values of each
    assert da1.mean().item() == 0.01444357167929411
    assert da2.mean().item() == 0.008129739202558994


def test_geopotential_approximation(met_cocip1: MetDataset):
    """Compare tau cirrus outputs when geopotential height is replaced by geometric height."""
    tc1 = tau_cirrus(met_cocip1)
    del met_cocip1.data["geopotential"]

    # Replace geopotential height with geometric height
    # This only has a "level" dimension, so this test also ensures that
    # the tau_cirrus function can handle this case.
    met_cocip1["geopotential_height"] = met_cocip1["altitude"]
    assert met_cocip1.data["geopotential_height"].dims == ("level",)
    tc2 = tau_cirrus(met_cocip1)

    assert tc1.dims == tc2.dims
    assert tc1.sizes == tc2.sizes

    # The important assertion is that the values are close enough from the perspective
    # of CoCiP.
    xr.testing.assert_allclose(tc1, tc2, rtol=0.04)



"""Test MetDataset and MetDataArrays are cached correctly when calling `__getitem__`."""

import pytest
import xarray as xr

from pycontrails import MetDataArray, MetDataset


def test_met_copy_data_arg(met_ecmwf_pl_path: str):
    """Test the `copy` argument in the MetDataset constructor."""
    ds = xr.open_dataset(met_ecmwf_pl_path)

    # The file on disk is not sorted correctly
    with pytest.raises(ValueError, match="not sorted"):
        MetDataset(ds, copy=False)

    # Sort it, then grab the data
    ds = MetDataset(ds).data

    # And test the `copy` parameter
    met = MetDataset(ds, copy=False)
    assert met.data is ds

    # Different objects if copied
    met = MetDataset(ds)
    assert met.data is not ds


def test_met_no_copy_same_view(met_ecmwf_pl_path: str):
    """Check that the `copy` flag provides view of same underlying data."""
    ds = xr.open_dataset(met_ecmwf_pl_path)

    # xarray __getitem__ counterintuitive
    assert ds["t"] is not ds["t"]
    assert ds["t"].values is not ds["t"].values

    # But they share the same view (so no memory leak)
    assert ds["t"].values.base is not None
    assert ds["t"].values.base is ds["t"].values.base

    # Old implementation is not memory efficient
    # Data is copied in the _preprocess_dims method
    mda1 = MetDataArray(ds["t"])
    mda2 = MetDataArray(ds["t"])
    assert mda1.data.values.base is not None
    assert mda2.data.values.base is not None
    assert mda1.data.values.base is not mda2.data.values.base

    # New implementation avoids this
    mda3 = MetDataArray(mda1.data, copy=False)
    mda4 = MetDataArray(mda1.data, copy=False)
    assert mda3.data is mda4.data
    assert mda3.data.values is mda4.data.values
    assert mda3.data.values.base is not None
    assert mda4.data.values.base is not None
    assert mda3.data.values.base is mda4.data.values.base
    assert mda1.data.values.base is mda3.data.values.base

    # And the culprit is _preprocess_dims
    mda3._preprocess_dims(False)
    mda4._preprocess_dims(False)
    assert mda3.data is not mda4.data
    assert mda3.data.values is not mda4.data.values



"""Test Fixtures.

Create test data for pycontrails.

See output in `pycontrails/tests/static`
"""

# import logging
# logger = logging.getLogger("pycontrails")
# logging.basicConfig(level=logging.DEBUG)

import pathlib
import sys
from datetime import datetime

import xarray as xr

from pycontrails import DiskCacheStore
from pycontrails.datalib.ecmwf import ERA5

# use separate disk cache
disk_cache = DiskCacheStore(cache_dir="cache")

# paths
parent = pathlib.Path(sys.argv[0]).parents[1].absolute()
static_dir = parent / "unit" / "static"

# domain
times = (datetime(2019, 5, 31, 5, 0, 0), datetime(2019, 5, 31, 6, 0, 0))
pl_variables = ["air_temperature", "specific_humidity", "specific_cloud_ice_water_content"]
sl_variables = ["surface_air_pressure"]
pressure_levels = [300, 250, 225]

### ERA5 PL data

era5pl = ERA5(
    times,
    variables=pl_variables,
    pressure_levels=pressure_levels,
    cachestore=disk_cache,
)
_ = era5pl.open_metdataset(xr_kwargs={"parallel": False})

# NOTE we are saving the cached CDS *source* files here.
# These are not the same as purely raw downloaded files from CDS.
ds_pl = xr.open_mfdataset(era5pl._cachepaths)
step = 100

ds = ds_pl.reindex(
    latitude=ds_pl["latitude"].values[0::step], longitude=ds_pl["longitude"].values[0::step]
)
ds.to_netcdf(static_dir / "met-ecmwf-pl.nc")


### ERA5 Single Level
era5sl = ERA5(times, variables=sl_variables, cachestore=disk_cache)
_ = era5sl.open_metdataset(xr_kwargs={"parallel": False})

# NOTE we are saving the raw CDS *source* files here
ds_sl = xr.open_mfdataset(era5sl._cachepaths)
step = 100
ds = ds_sl.reindex(
    latitude=ds_sl["latitude"].values[0::step], longitude=ds_sl["longitude"].values[0::step]
)
ds.to_netcdf(static_dir / "met-ecmwf-sl.nc")



"""Core data structures and methods."""

from pycontrails.core.cache import DiskCacheStore, GCPCacheStore
from pycontrails.core.datalib import MetDataSource
from pycontrails.core.fleet import Fleet
from pycontrails.core.flight import Aircraft, Flight
from pycontrails.core.fuel import Fuel, HydrogenFuel, JetA, SAFBlend
from pycontrails.core.met import MetDataArray, MetDataset
from pycontrails.core.met_var import MetVariable
from pycontrails.core.models import Model, ModelParams
from pycontrails.core.vector import GeoVectorDataset, VectorDataset

__all__ = [
    "Aircraft",
    "DiskCacheStore",
    "Fleet",
    "Flight",
    "Fuel",
    "GCPCacheStore",
    "GeoVectorDataset",
    "HydrogenFuel",
    "JetA",
    "MetDataArray",
    "MetDataset",
    "MetDataSource",
    "MetVariable",
    "Model",
    "ModelParams",
    "SAFBlend",
    "VectorDataset",
]



"""Load fixtures and paths corresponding to files in static directory."""

from __future__ import annotations

import json
import pathlib
from datetime import datetime

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from pycontrails import Aircraft, DiskCacheStore, Flight, MetDataArray, MetDataset
from pycontrails.core import cache, met, met_var
from pycontrails.datalib.ecmwf import ERA5
from pycontrails.models.aircraft_performance import AircraftPerformance
from tests import BADA3_PATH, BADA4_PATH, BADA_AVAILABLE

# find default cache dir for testing
DISK_CACHE_DIR = cache._get_user_cache_dir()

################################
# Fixtures in "static" directory
################################


def get_static_path(filename: str | pathlib.Path) -> pathlib.Path:
    """Return a path to file in ``/tests/static/`` directory.

    Parameters
    ----------
    filename : str | pathlib.Path
        Filename to prefix

    Returns
    -------
    pathlib.Path
    """
    parent = pathlib.Path(__file__).parent
    return parent / "static" / filename


@pytest.fixture(scope="session")
def flight_data() -> pd.DataFrame:
    """Load test flight data from csv.

    Scoped for the session.

    Returns
    -------
    pd.DataFrame
    """
    _path = get_static_path("flight.csv")
    parse_dates = ["time"]
    return pd.read_csv(_path, parse_dates=parse_dates)


@pytest.fixture(scope="session")
def flight_attrs() -> dict:
    """Load flight metadata from json.

    Scoped for the session.

    Returns
    -------
    dict
    """
    _path = get_static_path("flight-metadata.json")
    with open(_path) as f:
        return json.load(f)


@pytest.fixture(scope="session")
def override_cache() -> DiskCacheStore:
    """Override cache store while testing so cached met is not found.

    Returns
    -------
    DiskCacheStore
    """
    cache_dir = pathlib.Path(DISK_CACHE_DIR) / "test"
    return DiskCacheStore(cache_dir=cache_dir, allow_clear=True)


@pytest.fixture(scope="session")
def met_pcc_pl(met_ecmwf_pl_path: str, override_cache: DiskCacheStore) -> MetDataset:
    """Met data (pressure levels) for PCC algorithm testing.

    Returns
    -------
    MetDataset
    """

    times = (datetime(2019, 5, 31, 5, 0, 0), datetime(2019, 5, 31, 6, 0, 0))
    pl_variables = ["air_temperature", "specific_humidity", "specific_cloud_ice_water_content"]
    pressure_levels = [300, 250, 225]

    # load
    era5 = ERA5(
        time=times,
        variables=pl_variables,
        pressure_levels=pressure_levels,
        paths=met_ecmwf_pl_path,
        cachestore=override_cache,
    )
    return era5.open_metdataset()


@pytest.fixture(scope="session")
def met_pcc_sl(met_ecmwf_sl_path: str, override_cache: DiskCacheStore) -> MetDataset:
    """Met data (single level) for PCC algorithm testing.

    Returns
    -------
    MetDataset
    """
    times = (datetime(2019, 5, 31, 5, 0, 0), datetime(2019, 5, 31, 6, 0, 0))
    sl_variables = ["surface_air_pressure"]

    # load
    era5 = ERA5(
        time=times,
        variables=sl_variables,
        paths=met_ecmwf_sl_path,
        cachestore=override_cache,
    )
    return era5.open_metdataset()


@pytest.fixture(scope="session")
def met_accf_pl() -> MetDataset:
    """Met data (pressure levels) for PCC algorithm testing.

    Returns
    -------
    MetDataset
    """
    _path = get_static_path("met-accf-pl.nc")
    return MetDataset(xr.open_dataset(_path))


@pytest.fixture(scope="session")
def met_accf_sl() -> MetDataset:
    """Met data (single level) for PCC algorithm testing.

    Returns
    -------
    MetDataset
    """
    _path = get_static_path("met-accf-sl.nc")
    ds = xr.open_dataset(_path)
    ds = ds.expand_dims("level").assign_coords(level=("level", [-1]))
    return MetDataset(ds)


@pytest.fixture(scope="session")
def met_ecmwf_pl_path() -> str:
    """Path to ERA5 data at pressure levels.

    Returns
    -------
    str
    """
    _path = get_static_path("met-ecmwf-pl.nc")
    return str(_path)


@pytest.fixture(scope="session")
def met_ecmwf_sl_path() -> str:
    """Path to ERA5 data at single levels.

    Returns
    -------
    str
    """
    _path = get_static_path("met-ecmwf-sl.nc")
    return str(_path)


@pytest.fixture(scope="session")
def met_issr(met_ecmwf_pl_path: str) -> MetDataset:
    """ISSR output as MetDataset.

    As of pycontrails v0.42.0, there is an issue with dask threading causing
    the test `tests/unit/test_dtypes::test_issr_sac_grid_output[linear-ISSR]`
    to hang. As a workaround, we open without dask parallelism.

    Returns
    -------
    MetDataset
    """
    ds = xr.open_dataset(met_ecmwf_pl_path)
    ds = met.standardize_variables(ds, (met_var.AirTemperature, met_var.SpecificHumidity))
    ds.attrs["met_source"] = "ERA5"
    return MetDataset(ds)


@pytest.fixture
def met_cocip1() -> MetDataset:
    """ERA5 Meteorology to run cocip with ``flight_cocip1``.

    See ``met_cocip1_module_scope`` for a module-scoped version.

    See tests/fixtures/cocip-met.py for domain and creation of the source data file.
    """
    _path = get_static_path("met-era5-cocip1.nc")
    return MetDataset(xr.open_dataset(_path))


@pytest.fixture
def rad_cocip1() -> MetDataset:
    """ERA5 radiation data to run cocip with ``flight_cocip1``.

    See ``rad_cocip1_module_scope`` for a module-scoped version.

    See tests/fixtures/cocip-met.py for domain and creation of the source data file.
    """
    _path = get_static_path("rad-era5-cocip1.nc")
    return MetDataset(xr.open_dataset(_path))


@pytest.fixture(scope="module")
def met_cocip1_module_scope() -> MetDataset:
    """Create met available at module scope."""
    _path = get_static_path("met-era5-cocip1.nc")
    return MetDataset(xr.open_dataset(_path))


@pytest.fixture(scope="module")
def rad_cocip1_module_scope() -> MetDataset:
    """Create rad available at module scope."""
    _path = get_static_path("rad-era5-cocip1.nc")
    return MetDataset(xr.open_dataset(_path))


@pytest.fixture
def met_cocip2() -> MetDataset:
    """ERA5 meteorology to run cocip with ``flight-cocip2.csv``."""
    _path = get_static_path("met-era5-cocip2.nc")
    return MetDataset(xr.open_dataset(_path))


@pytest.fixture
def rad_cocip2() -> MetDataset:
    """ERA5 radiation data to run cocip with ``flight-cocip2.csv``."""
    _path = get_static_path("rad-era5-cocip2.nc")
    return MetDataset(xr.open_dataset(_path))


@pytest.fixture
def met_gfs() -> MetDataset:
    """
    Load GFS met example.

    See tests/fixtures/gfs-met.py for domain and creation of the source data file.
    """
    _path = get_static_path("met-gfs.nc")
    return MetDataset(xr.open_dataset(_path))


@pytest.fixture
def rad_gfs() -> MetDataset:
    """
    Load GFS rad example.

    See tests/fixtures/gfs-met.py for domain and creation of the source data file.
    """
    _path = get_static_path("rad-gfs.nc")
    return MetDataset(xr.open_dataset(_path))


@pytest.fixture(scope="function")  # keep function scoped
def met_era5_fake() -> MetDataset:
    """Create a fake ERA5 MetDataset."""
    shape = (360, 181, 7, 4)

    temp = 230 * np.ones(shape)
    sh = np.zeros(shape)
    for i in range(0, shape[0], 5):
        sh[i, :, :, :] = 1e-3

    ds = xr.Dataset(
        data_vars={
            "air_temperature": (["longitude", "latitude", "level", "time"], temp),
            "specific_humidity": (["longitude", "latitude", "level", "time"], sh),
        },
        coords={
            "longitude": np.linspace(-180, 179, shape[0]),
            "latitude": np.linspace(-90, 90, shape[1]),
            "level": [150, 175, 200, 225, 250, 300, 350],
            "time": pd.date_range("2020 Jan 1 00:00", "2020 Jan 1 03:00", shape[3]),
        },
    )
    return MetDataset(ds)


@pytest.fixture
def flight_cocip1() -> Flight:
    """Keep at function scope (default)."""
    # demo synthetic flight
    flight_parameters = {
        "flight_id": "test",
        "thrust": 0.22,  # thrust
        "nvpm_ei_n": 1.897462e15,
    }
    aircraft = Aircraft(aircraft_type="A380", wingspan=48, n_engine=2)

    # Example flight
    df = pd.DataFrame()
    df["longitude"] = np.linspace(-21, -23, 20)
    df["latitude"] = np.linspace(55, 57, 20)
    df["altitude"] = np.linspace(11000, 11500, 20)
    df["engine_efficiency"] = np.linspace(0.34, 0.35, 20)  # ope
    df["fuel_flow"] = np.linspace(2.1, 2.4, 20)  # kg/s
    df["aircraft_mass"] = np.linspace(154445, 154345, 20)  # kg
    df["time"] = pd.date_range("2019-01-01T00:15:00", "2019-01-01T02:30:00", periods=20)
    return Flight(df, aircraft=aircraft, attrs=flight_parameters)


@pytest.fixture
def flight_cocip2() -> Flight:
    """Test flight for cocip outputs.

    Compatible with ``met_cocip2`` and ``rad_cocip2``.
    """
    _path = get_static_path("flight-cocip2.csv")
    parse_dates = ["time"]
    return Flight(pd.read_csv(_path, parse_dates=parse_dates))


@pytest.fixture(scope="session")
def flight_meridian() -> Flight:
    """Test flight that crosses the meridian.

    Returns
    -------
    pd.DataFrame
    """
    _path = get_static_path("flight-meridian.csv")
    parse_dates = ["time"]
    df = pd.read_csv(_path, parse_dates=parse_dates)
    return Flight(df)


@pytest.fixture(scope="function")  # keep function scoped
def flight_fake() -> Flight:
    """Fake Flight fixture."""
    n_waypoints = 500

    df = pd.DataFrame(
        {
            "longitude": np.linspace(-45, 143, n_waypoints),
            "latitude": np.linspace(-20, 80, n_waypoints),
            "altitude": np.linspace(4000, 12000, n_waypoints),
            "time": pd.date_range("2020 Jan 1 00:17", "2020 Jan 1 02:36", n_waypoints),
        }
    )

    return Flight(df, attrs=dict(destination="SLC", flight_id="abcde"))


@pytest.fixture
def bada_model() -> AircraftPerformance:
    """Construct generic BADAFlight model."""

    BADAFlight = pytest.importorskip("pycontrails.ext.bada").BADAFlight

    if not BADA_AVAILABLE:
        pytest.skip("BADA data not available")

    params = {
        "bada3_path": BADA3_PATH,
        "bada4_path": BADA4_PATH,
    }
    return BADAFlight(params=params)


@pytest.fixture
def polygon_bug() -> MetDataArray:
    """Read the polygon bug example."""
    _path = get_static_path("polygon-bug.nc")
    da = xr.open_dataarray(_path)
    return MetDataArray(da)



# cython: language_level=3
"""Cythonized routines for the PycontrailsRegularGridInterpolator.

This module expands upon the cython interpolation routines introduced in
scipy 1.10 in several ways:

- Includes "fast path" linear interpolation routines for 3D and 4D arrays. The
  scipy implementation is only for 2D arrays.
- The fast path routines are specialized for float and double arrays. The scipy
  implementation is only for double arrays.
"""

from libc.math cimport NAN
import numpy as np

cimport cython
cimport numpy as np

np.import_array()


# See https://cython.readthedocs.io/en/latest/src/userguide/fusedtypes.html
ctypedef fused floating:
    float
    double


@cython.wraparound(False)
@cython.boundscheck(False)
@cython.initializedcheck(False)
@cython.nonecheck(False)
def evaluate_linear_4d(
    const floating[:, :, :, :] values,
    const long[:, :] indices,
    const double[:, :] norm_distances,
    floating[:] out,
) -> np.ndarray:
    cdef:
        long n_points = indices.shape[1]
        long i0, i1, i2, i3, p
        double y0, y1, y2, y3

    for p in range(n_points):
        i0 = indices[0, p]
        i1 = indices[1, p]
        i2 = indices[2, p]
        i3 = indices[3, p]

        if i0 < 0 or i1 < 0 or i2 < 0 or i3 < 0:
            out[p] = NAN
            continue

        y0 = norm_distances[0, p]
        y1 = norm_distances[1, p]
        y2 = norm_distances[2, p]
        y3 = norm_distances[3, p]

        out[p] = (
            values[i0, i1, i2, i3] * (1 - y0) * (1 - y1) * (1 - y2) * (1 - y3) +
            values[i0, i1, i2, i3+1] * (1 - y0) * (1 - y1) * (1 - y2) * y3 +
            values[i0, i1, i2+1, i3] * (1 - y0) * (1 - y1) * y2 * (1 - y3) +
            values[i0, i1, i2+1, i3+1] * (1 - y0) * (1 - y1) * y2 * y3 +
            values[i0, i1+1, i2, i3] * (1 - y0) * y1 * (1 - y2) * (1 - y3) +
            values[i0, i1+1, i2, i3+1] * (1 - y0) * y1 * (1 - y2) * y3 +
            values[i0, i1+1, i2+1, i3] * (1 - y0) * y1 * y2 * (1 - y3) +
            values[i0, i1+1, i2+1, i3+1] * (1 - y0) * y1 * y2 * y3 +
            values[i0+1, i1, i2, i3] * y0 * (1 - y1) * (1 - y2) * (1 - y3) +
            values[i0+1, i1, i2, i3+1] * y0 * (1 - y1) * (1 - y2) * y3 +
            values[i0+1, i1, i2+1, i3] * y0 * (1 - y1) * y2 * (1 - y3) +
            values[i0+1, i1, i2+1, i3+1] * y0 * (1 - y1) * y2 * y3 +
            values[i0+1, i1+1, i2, i3] * y0 * y1 * (1 - y2) * (1 - y3) +
            values[i0+1, i1+1, i2, i3+1] * y0 * y1 * (1 - y2) * y3 +
            values[i0+1, i1+1, i2+1, i3] * y0 * y1 * y2 * (1 - y3) +
            values[i0+1, i1+1, i2+1, i3+1] * y0 * y1 * y2 * y3
        )

    return np.asarray(out)



@cython.wraparound(False)
@cython.boundscheck(False)
@cython.initializedcheck(False)
@cython.nonecheck(False)
def evaluate_linear_3d(
    const floating[:, :, :] values,
    const long[:, :] indices,
    const double[:, :] norm_distances,
    floating[:] out,
) -> np.ndarray:
    cdef:
        long n_points = indices.shape[1]
        long i0, i1, i2, p
        double y0, y1, y2

    for p in range(n_points):
        i0 = indices[0, p]
        i1 = indices[1, p]
        i2 = indices[2, p]

        if i0 < 0 or i1 < 0 or i2 < 0:
            out[p] = NAN
            continue

        y0 = norm_distances[0, p]
        y1 = norm_distances[1, p]
        y2 = norm_distances[2, p]

        out[p] = (
            values[i0, i1, i2] * (1 - y0) * (1 - y1) * (1 - y2) +
            values[i0, i1, i2+1] * (1 - y0) * (1 - y1) * y2 +
            values[i0, i1+1, i2] * (1 - y0) * y1 * (1 - y2) +
            values[i0, i1+1, i2+1] * (1 - y0) * y1 * y2 +
            values[i0+1, i1, i2] * y0 * (1 - y1) * (1 - y2) +
            values[i0+1, i1, i2+1] * y0 * (1 - y1) * y2 +
            values[i0+1, i1+1, i2] * y0 * y1 * (1 - y2) +
            values[i0+1, i1+1, i2+1] * y0 * y1 * y2
        )

    return np.asarray(out)


@cython.wraparound(False)
@cython.boundscheck(False)
@cython.initializedcheck(False)
@cython.nonecheck(False)
def evaluate_linear_2d(
    const floating[:, :] values,
    const long[:, :] indices,
    const double[:, :] norm_distances,
    floating[:] out,
) -> np.ndarray:
    cdef:
        long n_points = indices.shape[1]
        long i0, i1, p
        double y0, y1

    for p in range(n_points):
        i0 = indices[0, p]
        i1 = indices[1, p]

        if i0 < 0 or i1 < 0:
            out[p] = NAN
            continue

        y0 = norm_distances[0, p]
        y1 = norm_distances[1, p]

        out[p] = (
            values[i0, i1] * (1 - y0) * (1 - y1) +
            values[i0, i1+1] * (1 - y0) * y1 +
            values[i0+1, i1] * y0 * (1 - y1) +
            values[i0+1, i1+1] * y0 * y1
        )

    return np.asarray(out
)


@cython.wraparound(False)
@cython.boundscheck(False)
@cython.initializedcheck(False)
@cython.nonecheck(False)
def evaluate_linear_1d(
    const floating[:] values,
    const long[:] indices,
    const double[:] norm_distances,
    floating[:] out,
) -> np.ndarray:
    cdef:
        long n_points = indices.shape[0]
        long i0, p
        double y0

    for p in range(n_points):
        i0 = indices[p]
        if i0 < 0:
            out[p] = NAN
            continue

        y0 = norm_distances[p]
        out[p] = values[i0] * (1 - y0) + values[i0+1] * y0

    return np.asarray(out)



"""Test contrail_grid module."""

from __future__ import annotations

from typing import Any, Iterable

import numpy as np
import pytest
import xarray as xr

try:
    from pycontrails.models.cocipgrid import CocipGrid
except ImportError:
    pytest.skip("CocipGrid not available", allow_module_level=True)

from pycontrails import GeoVectorDataset, MetDataset
from pycontrails.models.aircraft_performance import AircraftPerformance
from pycontrails.models.cocip import Cocip
from pycontrails.models.cocipgrid import cocip_grid as cg_module
from pycontrails.models.humidity_scaling import ExponentialBoostHumidityScaling
from pycontrails.utils.synthetic_flight import SyntheticFlight
from tests import BADA3_PATH, BADA4_PATH, BADA_AVAILABLE


@pytest.fixture
def source() -> MetDataset:
    """Return common source for `CocipGrid` model evaluation."""
    return CocipGrid.create_source(
        # levels do NOT need to agree with underlying met_cocip1
        level=[220, 230, 240, 250],
        time=np.datetime64("2019-01-01"),
        # using only a small set of longitude and latitude to avoid evolving out of bounds
        # the latitude usually goes first -- keep the range there tiny
        longitude=np.linspace(-35, -25, 40),
        latitude=np.linspace(51, 57, 20),
    )


@pytest.fixture
def instance_params(met_cocip1: MetDataset, rad_cocip1: MetDataset) -> dict[str, Any]:
    """Return common parameters for `CocipGrid` model instances.

    Note the fixture is scoped "function".
    """
    # met_cocip1 has levels [200, 225, 250, 300]
    np.testing.assert_array_equal(met_cocip1.data["level"], [200, 225, 250, 300])

    return {
        "met": met_cocip1,
        "rad": rad_cocip1,
        "dt_integration": np.timedelta64(5, "m"),
        # need to keep this super small to avoid advecting out of bounds
        "max_age": np.timedelta64(90, "m"),
        # keep low to ensure update_met_slices actually gets called
        "met_slice_dt": np.timedelta64(1, "h"),
        "bada3_path": BADA3_PATH,
        "bada4_path": BADA4_PATH,
        # explicitly raise error if we advect too far
        "interpolation_bounds_error": True,
        "target_split_size": 1000,
        "target_split_size_pre_SAC_boost": 1,
        "humidity_scaling": ExponentialBoostHumidityScaling(rhi_adj=0.9),
    }


def test_init_contrail_grid_minimal_params(met_cocip1: MetDataset, rad_cocip1: MetDataset) -> None:
    """Check that `CocipGrid` can be instantiated with a minimal set of parameters."""
    gc = CocipGrid(
        met=met_cocip1,
        rad=rad_cocip1,
        max_age=np.timedelta64(8, "m"),
        humidity_scaling=ExponentialBoostHumidityScaling(),
    )

    assert isinstance(gc.params, dict)
    assert isinstance(gc.interp_kwargs, dict)

    assert "dt_integration" in gc.params
    assert "max_age" in gc.params
    assert "aircraft_type" in gc.params
    assert "azimuth" in gc.params

    assert gc.params["met_slice_dt"] == np.timedelta64(1, "h")


def test_met_too_short(met_cocip1: MetDataset, rad_cocip1: MetDataset, source: MetDataset) -> None:
    """Check that a warning is issued when met and rad don't adequately overlap the source."""

    gc = CocipGrid(
        met=met_cocip1,
        rad=rad_cocip1,
        max_age=np.timedelta64(10, "h"),
        humidity_scaling=ExponentialBoostHumidityScaling(),
    )

    source = CocipGrid.create_source(level=[1, 2, 3], time=np.datetime64("2019-01-01"))
    gc.set_source(source)
    with pytest.warns(UserWarning, match="does not overlap the grid domain"):
        gc._check_met_source_overlap()

    source = CocipGrid.create_source(level=5, time=np.datetime64("2019-01-01T04"))
    gc.set_source(source)
    with pytest.warns(UserWarning, match="too short in the time dimension"):
        gc.attach_timedict()

    source = CocipGrid.create_source(level=5, time=np.datetime64("2019-01-01"))
    gc.set_source(source)
    with pytest.warns(UserWarning, match="does not overlap the grid domain"):
        gc._check_met_source_overlap()


@pytest.mark.parametrize("met_slice_dt", [np.timedelta64(1, "h"), np.timedelta64(2, "h")])
def test_load_met_slices(
    instance_params: dict[str, Any],
    source: MetDataset,
    met_slice_dt: np.timedelta64,
) -> None:
    """Verify expected behavior in `load_met_slices` method."""
    instance_params["met_slice_dt"] = met_slice_dt
    gc = CocipGrid(**instance_params)
    assert gc.met.shape == (16, 8, 4, 13)
    assert gc.rad.shape == (16, 8, 1, 13)

    (time,) = source.data["time"].values
    met, rad = gc._load_met_slices(time)

    if met_slice_dt == np.timedelta64(1, "h"):
        assert met.shape == (16, 8, 4, 2)
        assert rad.shape == (16, 8, 1, 3)  # rad shifted, so it includes 1 more time slice
    else:
        assert met.shape == (16, 8, 4, 3)
        assert rad.shape == (16, 8, 1, 4)  # rad shifted, so it includes 1 more time slice

    assert met.data["time"].values[0] == time
    assert rad.data["time"].values[0] == time - np.timedelta64(30, "m")

    # Make sure all variables present
    met.ensure_vars(gc.met_variables)


@pytest.mark.parametrize(
    "met_slice_dt",
    [
        np.timedelta64(0, "m"),
        np.timedelta64(55, "m"),
        np.timedelta64(150, "m"),
    ],
)
def test_init_bad_met_slice_dt(
    instance_params: dict[str, Any], met_slice_dt: np.timedelta64
) -> None:
    """Check that an error is raised when `met_slice_dt` is not a positive multiple of an hour."""
    assert instance_params["met_slice_dt"] == np.timedelta64(1, "h")
    instance_params["met_slice_dt"] = met_slice_dt

    if met_slice_dt == 0:
        match = "positive timedelta"
    else:
        match = "must be a multiple of the time difference"
    with pytest.raises(ValueError, match=match):
        CocipGrid(**instance_params)


@pytest.mark.parametrize(
    "dt_integration",
    [
        np.timedelta64(7, "m"),
        np.timedelta64(8, "m"),
        np.timedelta64(40, "m"),
    ],
)
def test_init_bad_dt_integration(
    instance_params: dict[str, Any], dt_integration: np.timedelta64
) -> None:
    """Check that an error is raised when dt_integration is not a positive multiple met_slice_dt."""
    assert instance_params["dt_integration"] == np.timedelta64(5, "m")
    instance_params["dt_integration"] = dt_integration
    with pytest.raises(ValueError, match="must be a multiple of dt_integration"):
        CocipGrid(**instance_params)


def test_create_bad_latitude() -> None:
    """Check that an error is raised when latitude values are too close to the poles."""
    with pytest.raises(ValueError, match="latitude"):
        CocipGrid.create_source(
            level=[220, 230, 240, 250],
            time=np.datetime64("2019-01-01"),
            longitude=np.linspace(-35, -25, 40),
            latitude=np.asarray([83, 85, 86]),
        )


def test_create_lon_lat_unspecified() -> None:
    """Ensure `CocipGrid` uses `lon_step` and `lat_step` of 1."""
    source = CocipGrid.create_source(level=[220, 250], time=np.datetime64("2019-01-01"))
    assert source.data["longitude"].size == 360
    assert source.data["latitude"].size == 161  # -80 to 80
    assert source.shape == (360, 161, 2, 1)


def test_create_lon_lat_step_args() -> None:
    """Ensure `CocipGrid` uses `lon_step` and `lat_step` specified in `__init__` args.."""
    source = CocipGrid.create_source(
        level=[220, 230, 240, 250],
        time=np.datetime64("2019-01-01"),
        lon_step=5,
        lat_step=5,
    )
    assert source.data["longitude"].size == 72
    assert source.data["latitude"].size == 33
    assert source.shape == (72, 33, 4, 1)


def test_init_avoid_double_process(instance_params: dict[str, Any]) -> None:
    """Check that CocipGrid doesn't alter met data twice."""
    # grab original
    met = instance_params["met"].copy()
    assert "tau_cirrus" not in met

    gc1 = CocipGrid(**instance_params)

    # The only change to met is the tau cirrus variable
    for var in met:
        np.testing.assert_array_equal(met[var].values, gc1.met[var].values)
    assert "tau_cirrus" not in met
    assert "tau_cirrus" in instance_params["met"]

    # grab a copy
    met1 = gc1.met.copy()
    rad1 = gc1.rad.copy()

    gc2 = CocipGrid(**instance_params)

    # Nothing has been tweaked the second time around
    assert gc1.met is gc2.met is instance_params["met"]
    assert gc1.rad is gc2.rad is instance_params["rad"]

    # The copied met and rad should not be the same object
    assert met1 is not gc2.met
    assert rad1 is not gc2.rad

    # But should contain the same values
    xr.testing.assert_equal(met1.data, gc2.met.data)
    xr.testing.assert_equal(rad1.data, gc2.rad.data)


@pytest.mark.parametrize("target_split_size", [100, 200, 300, 600, 1000, 10000])
def test_generate_new_grid_vectors(
    instance_params: dict[str, Any], target_split_size: int, source: MetDataset
) -> None:
    """Check that the `generate_new_vectors` method works as expected with MetDataset source."""
    instance_params["target_split_size"] = target_split_size
    gc = CocipGrid(**instance_params)
    gc.set_source(source)
    assert isinstance(gc.source_time, np.ndarray)
    assert gc.source_time.size == 1
    (time,) = gc.source_time
    time = time.astype("datetime64[h]")  # timedict uses this resolution

    gc.attach_timedict()
    assert time in gc.timedict
    filt = gc.timedict[time]

    vectors = gc._generate_new_vectors(filt)
    assert isinstance(vectors, Iterable)
    vectors = list(vectors)

    coords_size = (
        source.data["longitude"].size * source.data["latitude"].size * source.data["level"].size
    )
    vector_size = sum(vector.size for vector in vectors)
    assert coords_size == vector_size
    for vector in vectors:
        assert isinstance(vector, GeoVectorDataset)
        if target_split_size < coords_size:
            assert vector.size == pytest.approx(target_split_size, rel=0.5)
        else:
            assert len(vectors) == 1

    lon_concat = np.concatenate([vector["longitude"] for vector in vectors])
    lon_concat = np.unique(lon_concat)
    lon_concat.sort()
    np.testing.assert_array_almost_equal(lon_concat, source.data["longitude"].values, decimal=5)


##############################################################
# NOTE: No tests below here will run unless BADA is available.
##############################################################


@pytest.mark.skipif(not BADA_AVAILABLE, reason="BADA not available")
@pytest.mark.parametrize("aircraft_type", ["B737", "A320", "A359", "B772"])
@pytest.mark.parametrize("bada_priority", [3, 4])
def test_calc_emissions(
    instance_params: dict[str, Any],
    aircraft_type: str,
    bada_priority: int,
    source: MetDataset,
) -> None:
    """Test `calc_emissions` function."""
    instance_params["aircraft_type"] = aircraft_type
    # must be larger than 3200 so that all meshes are together
    instance_params["target_split_size"] = 5000
    instance_params["bada_priority"] = bada_priority

    gc = CocipGrid(**instance_params)
    gc.set_source(source)
    filt = np.array([True])

    vector = next(gc._generate_new_vectors(filt))

    met = instance_params["met"]
    # air_temperature need for calc_emissions
    vector["air_temperature"] = vector.intersect_met(met["air_temperature"], bounds_error=True)

    cg_module.calc_emissions(vector, gc.params)
    keys = [
        "air_temperature",
        "true_airspeed",
        "engine_efficiency",
        "fuel_flow",
        "nvpm_ei_n",
        "head_tail_dt",
    ]
    vector.ensure_vars(keys)

    assert "wingspan" in gc.params
    assert "aircraft_mass" in gc.params
    if bada_priority == 3:
        assert gc.params["bada_model"] == "BADA3"
    else:
        assert gc.params["bada_model"] == "BADA4"

    # Ensure not seeing crazy fluctuations in true_airspeed and fuel_flow
    # AND check that values are also non-constant
    tas = vector["true_airspeed"]
    ff = vector["fuel_flow"]
    assert len(np.unique(tas)) > 1
    assert len(np.unique(ff)) > 1

    assert 200 < np.mean(tas) < 300
    assert 0.5 < np.mean(ff) < 2.5

    assert np.std(tas) < 2
    assert np.std(ff) < 0.05


@pytest.mark.skipif(not BADA_AVAILABLE, reason="BADA not available")
def test_calc_first_contrail(instance_params: dict[str, Any], source: MetDataset) -> None:
    """Test `calc_first_contrail` method, which is called early in in `eval`."""
    instance_params["target_split_size"] = 5000
    gc = CocipGrid(**instance_params)
    gc.set_source(source)
    filt = np.array([True])

    vector = next(gc._generate_new_vectors(filt))
    assert vector.size == 3200

    met, rad = gc._load_met_slices(gc.source_time[0])

    cg_module.run_interpolators(
        vector,
        met,
        rad,
        dz_m=gc.params["dz_m"],
        humidity_scaling=gc.params["humidity_scaling"],
    )
    cg_module.calc_emissions(vector, gc.params)
    sac_vector = cg_module.find_initial_contrail_regions(vector, gc.params)[0]
    assert "T_crit_sac" in sac_vector

    cg_module.run_interpolators(
        sac_vector,
        met,
        rad,
        dz_m=gc.params["dz_m"],
        humidity_scaling=gc.params["humidity_scaling"],
    )
    contrail = cg_module.simulate_wake_vortex_downwash(sac_vector, gc.params)

    cg_module.run_interpolators(
        contrail,
        met,
        rad,
        dz_m=gc.params["dz_m"],
        humidity_scaling=gc.params["humidity_scaling"],
    )
    contrail = cg_module.find_initial_persistent_contrails(sac_vector, contrail, gc.params)[0]

    assert isinstance(contrail, GeoVectorDataset)

    assert sac_vector.size == 3199  # barely anything cut down in SAC check
    assert contrail.size == 429  # grabbed after letting test fail once

    # and level is slighty higher due to level shifting in downwash
    original_level = vector["level"][contrail["index"]]
    new_level = contrail["level"]
    assert np.all(new_level > original_level + 1)

    # not too much higher
    assert np.all(new_level < original_level + 2)


@pytest.fixture
def grid_results(instance_params: dict[str, Any]) -> MetDataset:
    """Run `CocipGrid` on three distinct times."""
    t_step = np.timedelta64(20, "m")
    start_time = np.datetime64("2019-01-01")
    source = CocipGrid.create_source(
        level=[220, 230, 240, 250],
        time=np.arange(start_time, start_time + np.timedelta64(1, "h"), t_step),
        longitude=np.linspace(-35, -25, 40),
        latitude=np.linspace(51, 57, 20),
    )

    gc = CocipGrid(**instance_params)
    return gc.eval(source=source)


@pytest.mark.skipif(not BADA_AVAILABLE, reason="BADA not available")
def test_grid_results(grid_results: MetDataset) -> None:
    """Test `grid_results` fixture.

    It would be nice to split this test into smaller pieces, but we are limited by the
    scope of grid_results (which we only want to run once).
    """
    # Test basic `MetDataset` type properties of `grid_results`.
    assert isinstance(grid_results, MetDataset)
    assert (np.diff(grid_results.data["time"].values) == np.timedelta64(20, "m")).all()

    for variable in ["contrail_age", "ef_per_m"]:
        assert variable in grid_results.data.data_vars

    assert grid_results["ef_per_m"].data.shape == (40, 20, 4, 3)

    # Ensure values are not NaN, and each level contains some initially persistent contrails
    for variable in ["contrail_age", "ef_per_m"]:
        assert np.all(np.isfinite(grid_results[variable].data.values))

        # nonzero somewhere
        da = grid_results[variable].data
        assert np.any(da.values != 0)

    # Test each `MetDataArray`s in `grid_results` have similar binary structure.
    # convert contrail_age to n_steps_with_persistent
    assert grid_results.attrs["dt_integration"] == "5 minutes"
    assert grid_results.attrs["max_age"] == "90 minutes"
    assert grid_results.attrs["bada_model"] == "BADA4"
    assert grid_results.attrs["aircraft_type"] == "B737"
    assert grid_results.attrs["humidity_scaling_name"] == "exponential_boost"
    assert grid_results.attrs["humidity_scaling_formula"]
    assert grid_results.attrs["humidity_scaling_rhi_adj"] == 0.9
    assert grid_results.attrs["humidity_scaling_rhi_boost_exponent"] == 1.7

    persistent = grid_results["contrail_age"].data > 0
    assert persistent.sum() == 605

    # zero outside persistent
    assert np.all(grid_results["ef_per_m"].data.values[~persistent] == 0)

    # positive inside
    assert np.all(grid_results["ef_per_m"].data.values[persistent] > 0)

    # Fixed in version 0.25.0: max age is adhered to by `contrail_age`
    assert np.all(grid_results["contrail_age"].data.values <= 1.5)

    # Pin the distribution of contrail ages
    ages, counts = np.unique(grid_results["contrail_age"].values, return_counts=True)
    np.testing.assert_allclose(ages, [0, 1 / 12, 1.5])
    np.testing.assert_array_equal(counts, [8995, 2, 603])

    # Ensure a hand picked pinned value is realized.
    # This test is HARD to maintain. Delete if it gets too annoying.
    point = grid_results.data.isel(longitude=14, latitude=19, level=2, time=2)
    assert point["contrail_age"].item() == 1.5
    assert point["ef_per_m"].item() == pytest.approx(46576688, rel=1e-3)


@pytest.fixture
def grid_results_segment_free(
    instance_params: dict[str, Any],
    grid_results: MetDataset,
) -> tuple[MetDataset, MetDataset]:
    """Run `CocipGrid` on skeleton of `grid_results` in a segment-free mode."""
    source = MetDataset.from_coords(**grid_results.coords)
    instance_params["azimuth"] = None
    instance_params["segment_length"] = None
    gc = CocipGrid(**instance_params)
    return grid_results, gc.eval(source)


@pytest.mark.skipif(not BADA_AVAILABLE, reason="BADA not available")
def test_grid_results_segment_free(
    grid_results_segment_free: tuple[MetDataset, MetDataset],
) -> None:
    """Test the `grid_results_segment_free` fixture.

    This is really just a sanity check that the segment-free mode is working.
    """
    out1, out2 = grid_results_segment_free
    assert out1.shape == out2.shape
    assert out1.data.data_vars.keys() == out2.data.data_vars.keys()

    assert np.count_nonzero(out1["ef_per_m"].values) == pytest.approx(605, abs=1)
    assert np.count_nonzero(out2["ef_per_m"].values) == pytest.approx(621, abs=1)

    # Pin some values
    da1 = out1["ef_per_m"].data
    filt1 = da1 > 0
    assert da1.where(filt1).mean().item() == pytest.approx(39581584, rel=1e-3)

    # In segment-free mode (generally and here), the mean nonzero EF is slightly lower
    da2 = out2["ef_per_m"].data
    filt2 = da2 > 0
    assert da2.where(filt2).mean().item() == pytest.approx(32914018, rel=1e-3)

    # For this example, we can say something about the distribution of EFs
    filt = filt1 & filt2
    v1 = da1.values[filt]
    v2 = da2.values[filt]
    xr.testing.assert_equal(filt, filt1)
    assert np.all(v1 > 0.5 * v2)
    assert np.all(v1 < 1.9 * v2)


@pytest.fixture
def syn_fl(instance_params: dict[str, Any], source: MetDataset) -> SyntheticFlight:
    """Return synthetic flight."""
    t_start = source.data["time"].values[0]
    t_stop = t_start + np.timedelta64(120, "m")
    bounds = {
        "longitude": source.data["longitude"].values,
        "latitude": source.data["latitude"].values,
        "level": source.data["level"].values,
        "time": np.array([t_start, t_stop]),
    }
    return SyntheticFlight(
        seed=5,
        bada4_path=BADA4_PATH,
        bounds=bounds,
        aircraft_type=CocipGrid.default_params.aircraft_type,
        u_wind=instance_params["met"]["eastward_wind"],
        v_wind=instance_params["met"]["northward_wind"],
    )


@pytest.mark.skipif(not BADA_AVAILABLE, reason="BADA not available")
def test_reasonable_syn_fl(syn_fl: SyntheticFlight, met_cocip1) -> None:
    """Check that syn_fl fixture is reasonable."""
    for _ in range(100):
        fl = syn_fl()
        fl.segment_angle
        u_wind = fl.intersect_met(met_cocip1["eastward_wind"])
        v_wind = fl.intersect_met(met_cocip1["northward_wind"])
        tas = fl.segment_true_airspeed(u_wind=u_wind, v_wind=v_wind)

        assert np.isnan(tas[-1])
        assert np.all(tas[:-1] < 250)
        assert np.all(tas[:-1] > 220)


@pytest.mark.skipif(not BADA_AVAILABLE, reason="BADA not available")
def test_geovector_source(syn_fl: SyntheticFlight, instance_params: dict[str, Any]) -> None:
    """Test `CocipGrid`  with GeoVectorDataset source."""
    # Call synthetic flight many times to create some random data.
    fls = [syn_fl() for _ in range(100)]
    source = GeoVectorDataset.sum(fls)

    # Pin the size
    assert source.size == 2608

    # Run the model
    gc = CocipGrid(**instance_params)
    out = gc.eval(source)
    assert isinstance(out, GeoVectorDataset)
    assert out.size == source.size
    assert "contrail_age" in out
    assert "ef_per_m" in out

    persistent = out["contrail_age"] > np.timedelta64(0, "ns")
    assert persistent.sum() == 72

    # Contrail age and positive EF are 1-1
    assert np.all(out["ef_per_m"][persistent] > 0)
    assert np.all(out["ef_per_m"][~persistent] == 0)

    # Pin the mean EF
    assert out["ef_per_m"].mean().item() == pytest.approx(875854, abs=10)
    assert out["ef_per_m"][persistent].mean().item() == pytest.approx(31725403, abs=10)


@pytest.mark.filterwarnings("ignore:invalid value encountered in remainder")
@pytest.mark.skipif(not BADA_AVAILABLE, reason="BADA not available")
def test_grid_against_flight(
    syn_fl: SyntheticFlight,
    instance_params: dict[str, Any],
    met_cocip1: MetDataset,
    rad_cocip1: MetDataset,
    bada_model: AircraftPerformance,
) -> None:
    """Confirm some agreement between traditional CoCiP and gridded CoCiP.

    Only if a waypoint is contrail forming, not the EF generated by waypoint.
    """
    # common cocip params
    params = {
        "max_age": instance_params["max_age"],
        "dt_integration": instance_params["dt_integration"],
        "humidity_scaling": ExponentialBoostHumidityScaling(rhi_adj=0.7),
        "interpolation_bounds_error": True,
    }

    # create the aircraft performance model from BADAFlight
    from pycontrails.ext.bada import BADAFlight

    bada_model = BADAFlight(bada4_path=instance_params["bada4_path"])

    # expect some persistent contrails
    n_fls_with_contrails = 0

    for _ in range(10):
        fl = syn_fl()
        fl["azimuth"] = fl.segment_azimuth()
        fl["true_airspeed"] = fl.segment_true_airspeed(
            fl.intersect_met(met_cocip1["eastward_wind"]),
            fl.intersect_met(met_cocip1["northward_wind"]),
        )

        cocip = Cocip(met_cocip1, rad_cocip1, params, aircraft_performance=bada_model)
        cocip_fl = cocip.eval(fl)
        cocip_fl_state = cocip_fl["cocip"].astype(bool)

        cg = CocipGrid(
            met=met_cocip1,
            rad=rad_cocip1,
            **params,
            bada4_path=instance_params["bada4_path"],
            verbose_outputs_formation="specific_humidity",
            verbose_outputs_evolution=True,
        )
        grid_fl = cg.eval(source=fl)
        grid_fl_state = grid_fl["ef_per_m"] != 0

        # Specifically check for humidity issues -- this has cropped up in the past
        q1 = cocip.source["specific_humidity"]
        q2 = grid_fl["specific_humidity"]
        np.testing.assert_array_equal(q1, q2)
        if hasattr(cocip, "_downwash_contrail"):
            q1 = cocip._downwash_contrail["specific_humidity"]
            q2 = cg.contrail.groupby("index")["specific_humidity"].first()

            # There can be slight inconsistencies in initially persistent.
            # This happens on one of the 10 flights. The Cocip model
            # predicts one more initially persistent waypoint compared with
            # the CocipGrid model. This is why we have some logic
            # below to check the sizes.
            if q1.size != q2.size:
                assert q1.size == q2.size + 1
                # Reassign q1, removing the extra waypoint
                q1 = cocip._downwash_contrail.dataframe.set_index("waypoint").loc[
                    q2.index, "specific_humidity"
                ]
            np.testing.assert_allclose(q1, q2, rtol=1e-8, atol=1e-5)

        # And there are at most two waypoint that doesn't show consistent
        # This is caused by continuity conventions
        assert np.sum(cocip_fl_state != grid_fl_state) <= 2
        if cocip_fl_state.sum():
            n_fls_with_contrails += 1

    assert n_fls_with_contrails == 8  # and we saw lots of persistent contrails


@pytest.mark.skipif(not BADA_AVAILABLE, reason="BADA not available")
def test_verbose_outputs_sac(instance_params: dict[str, Any], source: MetDataset):
    """Confirm the verbose_outputs parameter works with "sac" variable."""
    instance_params["verbose_outputs_formation"] = ["sac", "sdr"]

    model = CocipGrid(**instance_params)
    with pytest.warns(UserWarning, match="verbose_outputs"):
        ds = model.eval(source=source).data

    assert model.params["verbose_outputs_formation"] == {"sac"}
    assert isinstance(ds, xr.Dataset)

    assert "sac" in ds
    assert len(ds) == 3
    assert ds["sac"].sum() == 3199  # agrees with test_calc_first_contrail


@pytest.mark.skipif(not BADA_AVAILABLE, reason="BADA not available")
def test_verbose_outputs_formation(instance_params: dict[str, Any], source: MetDataset):
    """Confirm each verbose_outputs parameter is attached to results."""
    instance_params["verbose_outputs_formation"] = True
    model = CocipGrid(**instance_params)
    out = model.eval(source=source)

    expected = {
        "T_crit_sac",
        "persistent",
        "nvpm_ei_n",
        "sac",
        "engine_efficiency",
        "true_airspeed",
        "fuel_flow",
        "specific_humidity",
        "rhi",
        "iwc",
    }
    assert model.params["verbose_outputs_formation"] == expected

    assert len(out) == 10 + 2  # 10 verbose + 2 core (ef, age)
    for var in model.params["verbose_outputs_formation"]:
        assert var in out

    # Everything is a float
    ds = out.data
    for data in ds.values():
        assert data.dtype in ["float32", "float64"]

    # Pin a few values
    rel = 1e-4
    assert ds["T_crit_sac"].mean() == pytest.approx(222.54, rel=rel)
    assert ds["persistent"].sum() == 429
    assert (ds["contrail_age"] > 0).sum() == 235
    assert np.isfinite(ds["nvpm_ei_n"]).all()
    assert ds["nvpm_ei_n"].median() == pytest.approx(1.29718e15, rel=rel)
    assert ds["fuel_flow"].mean() == pytest.approx(0.6037, rel=rel)
    assert ds["rhi"].mean() == pytest.approx(0.6273, rel=rel)
    assert ds["iwc"].mean() == pytest.approx(4.08e-06, rel=rel)



"""Generate GFS test fixtures."""

import pathlib

from pycontrails import DiskCacheStore
from pycontrails.datalib.gfs import GFSForecast
from pycontrails.models import Cocip

# use separate disk cache
disk_cache = DiskCacheStore(cache_dir="cache")

# paths
parent = pathlib.Path(__file__).parents[1].absolute()
static_dir = parent / "unit" / "static"

# TODO: add test fixture for original GFS grib file
#


# see cocip_persistent test fixture
# this has to be > 2021 since older data is not available
times = ("2022-01-01 00:00:00", "2022-01-01 06:00:00")
met = GFSForecast(
    times,
    variables=Cocip.met_variables,
    pressure_levels=[200, 250, 300],
    cachestore=disk_cache,
    show_progress=True,
)
rad = GFSForecast(times, variables=Cocip.rad_variables, show_progress=True)

mds = met.open_metdataset(xr_kwargs=dict(parallel=False))
rds = rad.open_metdataset(xr_kwargs=dict(parallel=False))

for datasource, name in [(met, "met"), (rad, "rad")]:
    mds = datasource.open_metdataset(xr_kwargs=dict(parallel=False))

    # downselect region
    mds = mds.downselect([-40, 40, -20, 60])

    # reindex
    step = 5
    mds.data = mds.data.reindex(
        latitude=mds.data["latitude"].values[0::step],
        longitude=mds.data["longitude"].values[0::step],
    )
    mds.data.to_netcdf(static_dir / f"gfs-{name}.nc")



"""Contrail Cirrus Prediction (CoCiP) modeling support."""

from pycontrails.models.cocip.cocip import Cocip
from pycontrails.models.cocip.cocip_params import CocipFlightParams, CocipParams
from pycontrails.models.cocip.cocip_uncertainty import CocipUncertaintyParams

__all__ = [
    "Cocip",
    "CocipParams",
    "CocipUncertaintyParams",
    "CocipFlightParams",
]



"""Tools for creating synthetic data."""

from __future__ import annotations

import logging
import pathlib
import warnings
from typing import Any

import numpy as np
import pandas as pd
from pyproj.geod import Geod

from pycontrails.core.flight import Flight
from pycontrails.core.met import MetDataArray
from pycontrails.physics import constants, geo, units
from pycontrails.utils.types import ArrayOrFloat

try:
    from pycontrails.ext.bada import bada_model
except ImportError as e:
    raise ImportError(
        'SyntheticFlight requires BADA extension. Install with `pip install "pycontrails-bada @'
        ' git+ssh://git@github.com/contrailcirrus/pycontrails-bada.git"`'
    ) from e

logger = logging.getLogger(__name__)

SAMPLE_AIRCRAFT_TYPES = [
    "A20N",
    "A319",
    "A320",
    "A321",
    "A332",
    "A333",
    "A359",
    "A388",
    "B38M",
    "B737",
    "B738",
    "B739",
    "B744",
    "B752",
    "B763",
    "B772",
    "B77W",
    "B788",
    "B789",
    "CRJ2",
    "CRJ7",
    "CRJ9",
    "E145",
    "E190",
    "E195",
    "E75L",
    "E75S",
]


class SyntheticFlight:
    """Create a synthetic flight."""

    # Random number generator
    rng = np.random.default_rng(None)

    # Maximum queue size. When generating many thousands of flights, performance is slightly
    # improved if this parameter is increased.
    max_queue_size = 100

    # Minimum number of waypoints in flight
    min_n_waypoints = 10

    # Type hint a few instance variables
    bada: bada_model.BADA | None
    aircraft_type: str

    longitude_min: float
    longitude_max: float
    latitude_min: float
    latitude_max: float
    level_min: float
    level_max: float
    time_min: np.datetime64
    time_max: np.datetime64

    def __init__(
        self,
        bounds: dict[str, np.ndarray],
        aircraft_type: str | None = None,
        bada3_path: str | pathlib.Path | None = None,
        bada4_path: str | pathlib.Path | None = None,
        speed_m_per_s: float | None = None,
        seed: int | None = None,
        u_wind: MetDataArray | None = None,
        v_wind: MetDataArray | None = None,
    ) -> None:
        """Create a synthetic flight generator.

        Parameters
        ----------
        bounds : dict[str, np.ndarray]
            A dictionary with keys "longitude", "latitude", "level", and "time". All synthetic
            flights will have coordinates bounded by the extreme values in this dictionary.
        aircraft_type : str
            A flight type, assumed to exist in the BADA3 or BADA4 dataset.
            If None provided, a random aircraft type will be chosen from
            ``SAMPLE_AIRCRAFT`` on every call.
        bada3_path : str | pathlib.Path, optional
            A path to a local BADA3 data source.
            Defaults to None.
        bada4_path : str | pathlib.Path, optional
            A path to a local BADA4 data source.
            Defaults to None.
        speed_m_per_s : float, optional
            Directly define cruising air speed. Only used if `bada3_path` and `bada4_path` are
            not specified. By default None.
        seed : int, optional
            Reseed the random generator. By default None.
        u_wind, v_wind : MetDataArray, optional
            Eastward and northward wind data. If provided, flight true airspeed is computed
            with respect to wind.
        max_queue_size : int, optional
            Maximum queue size. When generating many thousands of flights, performance is slightly
            improved if this parameter is increased.
        min_n_waypoints : int, optional
            Minimum number of waypoints in flight
        """
        if seed is not None:
            self.rng = np.random.default_rng(seed)

        self.bounds = bounds
        self.constant_aircraft_type = aircraft_type
        self.bada3_path = bada3_path
        self.bada4_path = bada4_path
        self.speed_m_per_s = speed_m_per_s
        self.geod = Geod(a=constants.radius_earth)

        def extremes(arr: np.ndarray) -> tuple[Any, Any]:
            arr = np.asarray(arr)
            return arr.min(), arr.max()

        self.longitude_min, self.longitude_max = extremes(bounds["longitude"])
        self.latitude_min, self.latitude_max = extremes(bounds["latitude"])
        self.level_min, self.level_max = extremes(bounds["level"])
        self.time_min, self.time_max = extremes(bounds["time"])

        self.u_wind = u_wind
        self.v_wind = v_wind

        self._queue: list[Flight] = []
        self._depth = 0

    def __repr__(self) -> str:
        """Get string representation."""
        msg = "Synthetic flight generator with parameters:"
        for key in ["longitude", "latitude", "level", "time"]:
            msg += "\n"
            for bound in ["min", "max"]:
                attr = f"{key}_{bound}"
                val = getattr(self, attr)
                msg += f"{attr}: {val} "
        msg += f"\nmax_queue_size: {self.max_queue_size}  min_n_waypoints: {self.min_n_waypoints}"
        return msg

    def __call__(self, timestep: np.timedelta64 = np.timedelta64(1, "m")) -> Flight:
        """Create random flight within `bounds` at a constant altitude.

        BADA4 data is used to determine flight speed at a randomly chosen altitude within
        the level constraints defined by `bounds`. The flight trajectory follows a great circle
        from a uniformly random chosen source point to a uniformly randomly chosen destination.

        For generating a large number of random flights, it's best to call this within a
        generator expression, ie,
            ``fls = (syn() for _ in range(100_000))``

        Parameters
        ----------
        timestep : np.timedelta64, optional
            Time interval between waypoints. By default, 1 minute

        Returns
        -------
        Flight
            Random `Flight` instance constrained by bounds.
        """
        # Building flights with `u_wind` and `v_wind` involved in the true airspeed calculation is
        # slow. BUT, we can do it in a vectorized way. So we maintain a short queue that gets
        # repeatedly replenished.
        if self.u_wind is not None and self.v_wind is not None:
            while not self._queue:
                # Need to do some significant refactor to use generators seemlessly
                # The parameter timestep can change between calls
                new_batch = self._generate_with_wind(self.max_queue_size, timestep)
                self._queue.extend(new_batch)
            fl = self._queue.pop()
            while np.any(np.diff(fl["time"]) != timestep):
                logger.debug("Found flight in queue with bad timestep.")
                fl = self._queue.pop()
            return fl

        self._depth = 0
        self._define_aircraft()
        return self._generate_single_flight_no_wind(timestep)

    def _id(self) -> int:
        """Get random flight ID."""
        return self.rng.integers(100_000, 999_999)

    def _define_aircraft(self) -> None:
        """Define or update instance variables pertaining to flight aircrafts.

        Specify
            - aircraft_type
            - bada_enabled
            - fl_all
            - cruise_tas

        Raises
        ------
        FileNotFoundError
            BADA files not found despite non-default BADA3 or BADA4 paths
        ValueError
            BADA files not found under default paths AND speed_m_per_s not defined
        """
        if self.constant_aircraft_type is None:
            self.aircraft_type = self.rng.choice(SAMPLE_AIRCRAFT_TYPES)
        else:
            self.aircraft_type = self.constant_aircraft_type

        try:
            self.bada = bada_model.get_bada(
                self.aircraft_type,
                bada3_path=self.bada3_path,
                bada4_path=self.bada4_path,
                bada_priority=4,
            )

        except FileNotFoundError as err:
            logger.warning("BADA files not found")

            # If non-default bada paths were passed into __init__, we should raise an error
            if self.bada3_path is not None or self.bada4_path is not None:
                raise FileNotFoundError from err

            # If  bada paths were not passed into __init__, we expect to know speed_m_per_s
            if self.speed_m_per_s is None:
                raise ValueError("Either specify `bada3_path`, `bada4_path`, or `speed_m_per_s`.")
            self.bada = None

    def _calc_speed_m_per_s(self, level: ArrayOrFloat) -> ArrayOrFloat:
        if self.bada is not None:
            alt_ft = units.pl_to_ft(level)
            return self.bada.nominal_cruising_speed(self.aircraft_type, alt_ft)

        if self.speed_m_per_s is None:
            raise ValueError("Either specify `bada3_path`, `bada4_path` or `speed_m_per_s`.")

        if isinstance(level, np.ndarray):
            return np.full_like(level, self.speed_m_per_s)
        return self.speed_m_per_s

    def _generate_single_flight_no_wind(self, timestep: np.timedelta64) -> Flight:
        src_lon = self.rng.uniform(self.longitude_min, self.longitude_max)
        src_lat = self.rng.uniform(self.latitude_min, self.latitude_max)
        dest_lon = self.rng.uniform(self.longitude_min, self.longitude_max)
        dest_lat = self.rng.uniform(self.latitude_min, self.latitude_max)
        src = src_lon, src_lat
        dest = dest_lon, dest_lat
        az, _, dist = self.geod.inv(*src, *dest)

        level = self.rng.uniform(self.level_min, self.level_max)
        speed_m_per_s = self._calc_speed_m_per_s(level)

        m_per_timestep = speed_m_per_s * (timestep / np.timedelta64(1, "s"))
        npts = int(dist // m_per_timestep)  # need to cast: dist is np.float64

        # Dealing with situations of npts too small or too big
        if npts > (self.time_max - self.time_min) / timestep:
            msg = (
                "Not enough available time in `bounds` to create good flight between "
                f"source {src} and destination {dest}. Try enlarging the time dimension, "
                "or reducing the longitude and latitude dimensions."
            )

            new_npts = int((self.time_max - self.time_min) / timestep)
            logger.debug("Override npts from %s to %s", npts, new_npts)
            npts = new_npts

            if npts < self.min_n_waypoints:
                raise ValueError(msg)
            warnings.warn(msg)

        if npts < self.min_n_waypoints:
            # Try 10 times, then give up.
            self._depth += 1
            if self._depth > 10:
                raise ValueError("Cannot create flight. Increase dimensions in `bounds`.")
            return self._generate_single_flight_no_wind(timestep)  # recursive

        result = self.geod.fwd_intermediate(
            *src, az, npts, m_per_timestep, return_back_azimuth=False  # type: ignore
        )
        longitude = np.asarray(result.lons)
        latitude = np.asarray(result.lats)
        if geo.haversine(longitude[-1], latitude[-1], *dest) > m_per_timestep:
            logger.debug(
                "Synthetic flight did not arrive at destination. "
                "This is likely due to overriding npts."
            )

        rand_range = int((self.time_max - self.time_min - npts * timestep) / timestep) + 1
        time0 = self.time_min + self.rng.integers(rand_range) * timestep
        time: np.ndarray = np.arange(time0, time0 + npts * timestep, timestep)

        df = {"longitude": longitude, "latitude": latitude, "level": level, "time": time}
        return Flight(pd.DataFrame(df), flight_id=self._id(), aircraft_type=self.aircraft_type)

    def _generate_with_wind(self, n: int, timestep: np.timedelta64) -> list[Flight]:
        logger.debug("Generate %s new flights with wind", n)

        self._define_aircraft()

        # Step 1: Randomly select longitude, latitude, level
        src_lon = self.rng.uniform(self.longitude_min, self.longitude_max, n)
        src_lat = self.rng.uniform(self.latitude_min, self.latitude_max, n)
        dest_lon = self.rng.uniform(self.longitude_min, self.longitude_max, n)
        dest_lat = self.rng.uniform(self.latitude_min, self.latitude_max, n)
        level = self.rng.uniform(self.level_min, self.level_max, n)
        src = src_lon, src_lat
        dest = dest_lon, dest_lat
        az: np.ndarray
        dist: np.ndarray
        az, _, dist = self.geod.inv(*src, *dest)

        # Step 2: Compute approximate flight times according to nominal (no wind) TAS
        nom_speed_m_per_s = self._calc_speed_m_per_s(level)
        # NOTE: Because of casting, the multiplication below is NOT associative!
        # In other words, the parentheses are needed on the right-most term
        nom_m_per_timestep = nom_speed_m_per_s * (timestep / np.timedelta64(1, "s"))
        approx_flight_duration_s = dist / nom_speed_m_per_s * np.timedelta64(1, "s")

        # Step 3: Randomly select start time -- use 0.9 for a small buffer
        rand_float = 0.9 * self.rng.random(n)
        time_windows = self.time_max - self.time_min - approx_flight_duration_s

        # Here `time_windows` can have negative timedeltas, which is not good.
        n_negative = np.sum(time_windows < np.timedelta64(0, "s"))
        logger.debug(
            "Found %s / %s src-dist pairs not fitting into the time dimension", n_negative, n
        )
        if n_negative >= 0.1 * n:
            warnings.warn(
                "Not enough available time in `bounds` to create reasonable random flights. Try "
                "enlarging the time dimension, or reducing the longitude and latitude dimensions."
            )
        # Manually clip at 0
        time_windows = np.maximum(time_windows, np.timedelta64(0, "s"))

        rand_start = rand_float * time_windows
        start_time = self.time_min + rand_start

        # Step 4: Iterate and collect
        lons = []
        lats = []
        times = []

        def take_step(
            cur_lon: np.ndarray,
            cur_lat: np.ndarray,
            cur_time: np.typing.NDArray[np.datetime64],
            cur_az: np.ndarray,
            cur_active: np.ndarray,
        ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
            sin, cos = geo.azimuth_to_direction(cur_az, cur_lat)

            if self.u_wind is None or self.v_wind is None:
                raise TypeError("Define attributes `u_wind` and `v_wind`.")
            u_wind = self.u_wind.interpolate(cur_lon, cur_lat, level, cur_time)
            v_wind = self.v_wind.interpolate(cur_lon, cur_lat, level, cur_time)

            nom_x = nom_m_per_timestep * cos
            nom_y = nom_m_per_timestep * sin
            # NOTE: Because of casting, the multiplication below is NOT associative.
            # In other words, the parentheses are needed on the right-most term
            tas_x = nom_x + u_wind * (timestep / np.timedelta64(1, "s"))
            tas_y = nom_y + v_wind * (timestep / np.timedelta64(1, "s"))
            tas = (tas_x**2 + tas_y**2) ** 0.5

            next_lon, next_lat, _ = self.geod.fwd(cur_lon, cur_lat, cur_az, tas)
            next_az: np.ndarray
            next_dist: np.ndarray
            next_az, _, next_dist = self.geod.inv(next_lon, next_lat, *dest)
            next_time = cur_time + timestep
            next_active = (
                cur_active & (next_dist > nom_m_per_timestep) & (next_time < self.time_max)
            )

            return next_lon, next_lat, next_time, next_az, next_active

        lon = src_lon
        lat = src_lat
        time = start_time
        az = az
        active = np.ones_like(lon).astype(bool)

        while active.sum():
            lons.append(np.where(active, lon, np.nan))
            lats.append(np.where(active, lat, np.nan))
            times.append(time)
            lon, lat, time, az, active = take_step(lon, lat, time, az, active)

        # Step 5: Assemble and return
        lons_arr = np.asarray(lons).T
        lats_arr = np.asarray(lats).T
        times_arr = np.asarray(times).T
        data = [
            {"longitude": lon, "latitude": lat, "level": level, "time": time}
            for lon, lat, level, time in zip(lons_arr, lats_arr, level, times_arr)
        ]
        dfs = [pd.DataFrame(d).dropna() for d in data]
        dfs = [df for df in dfs if len(df) >= self.min_n_waypoints]
        dfs = [df.assign(altitude=units.pl_to_m(df["level"])).drop(columns="level") for df in dfs]

        return [Flight(df, flight_id=self._id(), aircraft_type=self.aircraft_type) for df in dfs]



"""Grid-like Contrail Cirrus Prediction (CoCiP) modeling support."""

from pycontrails.models.cocipgrid.cocip_grid import CocipGrid
from pycontrails.models.cocipgrid.cocip_grid_params import CocipGridParams

__all__ = [
    "CocipGrid",
    "CocipGridParams",
]



"""GFS Parameter Support.

Parameter definitions:

- `Reanalysis <https://www.nco.ncep.noaa.gov/pmb/products/gfs/gfs.t00z.pgrb2.0p25.anl.shtml>`_
- `Surface <https://www.nco.ncep.noaa.gov/pmb/products/gfs/gfs.t00z.pgrb2.0p25.f000.shtml>`_
- `Pressure Levels <https://www.nco.ncep.noaa.gov/pmb/products/gfs/gfs.t00z.pgrb2.0p25.f003.shtml>`_
"""

from __future__ import annotations

from pycontrails.core import met_var
from pycontrails.core.met import MetVariable

TotalCloudCoverIsobaric = MetVariable(
    short_name="tcc",
    standard_name="total_cloud_cover_isobaric",
    long_name="Total cloud cover at isobaric surface",
    level_type="isobaricInhPa",
    ecmwf_id=228164,
    grib2_id=(0, 6, 1),
    units="%",
    description=(
        "This parameter is the percentage of a grid box covered by cloud (liquid or ice) at a"
        " specific pressure level."
    ),
)

CloudIceWaterMixingRatio = MetVariable(
    short_name="icmr",
    standard_name="ice_water_mixing_ratio",
    long_name="Cloud ice water mixing ratio",
    units="kg kg**-1",
    level_type="isobaricInhPa",
    ecmwf_id=260019,
    grib2_id=(0, 1, 23),
    description=(
        "This parameter is the mass of cloud ice particles per kilogram of the total mass of dry"
        " air. "
    ),
)


TOAUpwardShortwaveRadiation = MetVariable(
    short_name="uswrf",
    standard_name="toa_upward_shortwave_flux",
    long_name="Top of atmosphere upward shortwave radiation",
    units="W m**-2",
    level_type="nominalTop",
    grib2_id=(0, 4, 193),
    description=(
        "This parameter is the outgoing shortwave (solar) radiation at the nominal top of the"
        " atmosphere."
    ),
)

TOAUpwardLongwaveRadiation = MetVariable(
    short_name="ulwrf",
    standard_name="toa_upward_longwave_flux",
    long_name="Top of atmosphere upward longwave radiation",
    units="W m**-2",
    level_type="nominalTop",
    grib2_id=(0, 5, 193),
    description=(
        "This parameter is the outgoing longwave (thermal) radiation at the nominal top of the"
        " atmosphere."
    ),
)

Visibility = MetVariable(
    short_name="vis",
    standard_name="visibility",
    long_name="Visibility at ground or water surface",
    units="m",
    level_type="surface",
    grib2_id=(0, 19, 0),
    description="This parameter is the visibility at the ground or water surface, in meters.",
)


PRESSURE_LEVEL_VARIABLES = [
    met_var.AirTemperature,
    met_var.SpecificHumidity,
    met_var.RelativeHumidity,
    met_var.GeopotentialHeight,
    met_var.EastwardWind,
    met_var.NorthwardWind,
    met_var.VerticalVelocity,
    CloudIceWaterMixingRatio,
    TotalCloudCoverIsobaric,
]
SURFACE_VARIABLES = [
    met_var.SurfacePressure,
    Visibility,
    TOAUpwardShortwaveRadiation,
    TOAUpwardLongwaveRadiation,
]

GFS_VARIABLES = PRESSURE_LEVEL_VARIABLES + SURFACE_VARIABLES



"""ECWMF IFS forecast data access."""

from __future__ import annotations

import logging
import pathlib
import warnings
from datetime import datetime
from typing import Any

LOG = logging.getLogger(__name__)

import numpy as np
import pandas as pd
import xarray as xr
from overrides import overrides

from pycontrails.core import datalib, met
from pycontrails.datalib.ecmwf.variables import ECMWF_VARIABLES
from pycontrails.physics import constants
from pycontrails.utils.types import DatetimeLike


class IFS(datalib.MetDataSource):
    """
    ECMWF Integrated Forecasting System (IFS) data source.

    .. warning::

        This data source is fully implemented.

    Parameters
    ----------
    time : datalib.TimeInput | None
        The time range for data retrieval, either a single datetime or (start, end) datetime range.
        Input must be a single datetime-like or tuple of datetime-like
        (datetime, :class:`pandas.Timestamp`, :class:`numpy.datetime64`)
        specifying the (start, end) of the date range, inclusive.
        If None, all time coordinates will be loaded.
    variables : datalib.VariableInput
        Variable name (i.e. "air_temperature", ["air_temperature, relative_humidity"])
        See :attr:`pressure_level_variables` for the list of available variables.
    pressure_levels : datalib.PressureLevelInput, optional
        Pressure level bounds for data (min, max), in hPa (mbar)
        Set to -1 for to download surface level parameters.
        Defaults to -1.
    paths : str | list[str] | pathlib.Path | list[pathlib.Path] | None, optional
        UNSUPPORTED FOR IFS
    forecast_path: str | pathlib.Path | None, optional
        Path to local forecast files.
        Defaults to None
    forecast_date: DatetimeLike, optional
        Forecast date to load specific netcdf files.
        Defaults to None

    Notes
    -----
    This takes an average pressure of the model level to create
    pressure level dimensions.
    """

    __slots__ = ("forecast_date", "forecast_path")

    #: Root path of IFS data
    forecast_path: pathlib.Path

    #: Forecast datetime of IFS forecast
    forecast_date: pd.Timestamp

    def __init__(
        self,
        time: datalib.TimeInput | None,
        variables: datalib.VariableInput,
        pressure_levels: datalib.PressureLevelInput = -1,
        paths: str | list[str] | pathlib.Path | list[pathlib.Path] | None = None,
        grid: float | None = None,
        forecast_path: str | pathlib.Path | None = None,
        forecast_date: DatetimeLike | None = None,
    ) -> None:
        self.paths = paths  # TODO: this is currently unused
        self.grid = grid  # TODO: this is currently unused

        # path to forecast files
        if forecast_path is None:
            raise ValueError("Forecast path input is required for IFS")
        self.forecast_path = pathlib.Path(forecast_path)

        # TODO: automatically select a forecast_date from input time range?
        self.forecast_date = pd.to_datetime(forecast_date).to_pydatetime()

        # parse inputs
        self.timesteps = datalib.parse_timesteps(time, freq="3H")
        self.pressure_levels = datalib.parse_pressure_levels(pressure_levels, None)
        self.variables = datalib.parse_variables(variables, self.supported_variables)

    def __repr__(self) -> str:
        base = super().__repr__()
        return f"{base}\n\tForecast date: {self.forecast_date}"

    @property
    def supported_variables(self) -> list[met.MetVariable]:
        """IFS parameters available.

        Returns
        -------
        list[MetVariable] | None
            List of MetVariable available in datasource
        """
        return ECMWF_VARIABLES

    @property
    def supported_pressure_levels(self) -> None:
        """IFS does not provide constant pressure levels and instead uses model levels.

        Returns
        -------
        list[int]
        """
        return None

    @overrides
    def open_metdataset(
        self,
        dataset: xr.Dataset | None = None,
        xr_kwargs: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> met.MetDataset:
        xr_kwargs = xr_kwargs or {}

        #  short-circuit dataset or file paths if provided
        if self.paths is not None or dataset is not None:
            raise NotImplementedError("IFS input paths or input dataset is not supported")

        # load / merge datasets
        ds = self._open_ifs_dataset(**xr_kwargs)

        # drop ancillary vars
        ds = ds.drop_vars(names=["hyai", "hybi"])

        # downselect dataset if only a subset of times, pressure levels, or variables are requested
        if self.timesteps:
            ds = ds.sel(time=self.timesteps)
        else:
            # set timesteps from dataset "time" coordinates
            # np.datetime64 doesn't covert to list[datetime] unless its unit is us
            self.timesteps = ds["time"].values.astype("datetime64[us]").tolist()

        # downselect hyam/hybm coefficients by the "lev" coordinate
        # (this is a 1-indexed verison of nhym)
        ds["hyam"] = ds["hyam"][dict(nhym=(ds["lev"] - 1).astype(int))]
        ds["hybm"] = ds["hybm"][dict(nhym=(ds["lev"] - 1).astype(int))]

        # calculate air_pressure (Pa) by hybrid sigma pressure
        ds["air_pressure"] = ds["hyam"] + (ds["hybm"] * ds["surface_pressure"])
        ds["air_pressure"].attrs["units"] = "Pa"
        ds["air_pressure"].attrs["long_name"] = "Air pressure"

        # calculate virtual temperature (t_virtual)
        # the temperature at which dry air would have the same density
        # as the moist air at a given pressure
        ds["t_virtual"] = ds["t"] * (1 + ds["q"] * ((constants.R_v / constants.R_d) - 1))
        ds["t_virtual"].attrs["units"] = "K"
        ds["t_virtual"].attrs["long_name"] = "Virtual Temperature"

        # calculate geopotential
        if "z" in self.variable_shortnames:
            ds["z"] = self._calc_geopotential(ds)

        # take the mean of the air pressure to create quasi-gridded level coordinate
        ds = ds.assign_coords(
            {"level": ("lev", (ds["air_pressure"].mean(dim=["time", "lat", "lon"]) / 100).values)}
        )
        ds = ds.swap_dims({"lev": "level"})
        ds = ds.drop_vars(names=["lev"])

        # rename dimensions
        ds = ds.rename({"lat": "latitude", "lon": "longitude"})

        # downselect variables
        ds = ds[self.variable_shortnames]

        # TODO: fix this correctly
        if "level" not in ds.dims:
            ds = ds.expand_dims({"level": [-1]})

        # harmonize variable names
        ds = met.standardize_variables(ds, self.variables)

        ds.attrs["met_source"] = type(self).__name__
        return met.MetDataset(ds, **kwargs)

    @overrides
    def download_dataset(self, times: list[datetime]) -> None:
        raise NotImplementedError("IFS download is not supported")

    @overrides
    def cache_dataset(self, dataset: xr.Dataset) -> None:
        raise NotImplementedError("IFS dataset caching not supported")

    @overrides
    def create_cachepath(self, t: datetime) -> str:
        raise NotImplementedError("IFS download is not supported")

    def _open_ifs_dataset(self, **xr_kwargs: Any) -> xr.Dataset:
        # get the path to each IFS file for each forecast date
        date_str = self.forecast_date.strftime("%Y%m%d")
        path_full = f"{self.forecast_path}/FC_{date_str}_00_144.nc"
        path_fl = f"{self.forecast_path}/FC_{date_str}_00_144_fl.nc"
        path_surface = f"{self.forecast_path}/FC_{date_str}_00_144_sur.nc"
        path_rad = f"{self.forecast_path}/FC_{date_str}_00_144_rad.nc"

        # load each dataset
        LOG.debug(f"Loading IFS forecast date {date_str}")

        # load each dataset
        xr_kwargs.setdefault("chunks", datalib.DEFAULT_CHUNKS)
        xr_kwargs.setdefault("engine", datalib.NETCDF_ENGINE)
        ds_full = xr.open_dataset(path_full, **xr_kwargs)
        ds_fl = xr.open_dataset(path_fl, **xr_kwargs)
        ds_surface = xr.open_dataset(path_surface, **xr_kwargs)
        ds_rad = xr.open_dataset(path_rad, **xr_kwargs)

        # calculate surface pressure from ln(surface pressure) var, squeeze out "lev" dim
        ds_full["surface_pressure"] = np.exp(ds_full["lnsp"]).squeeze()
        ds_full["surface_pressure"].attrs["units"] = "Pa"
        ds_full["surface_pressure"].attrs["long_name"] = "Surface air pressure"

        # swap dim names for consistency
        ds_full = ds_full.drop_vars(names=["lnsp", "lev"])
        ds_full = ds_full.rename({"lev_2": "lev"})

        # drop vars so all datasets can merge
        ds_fl = ds_fl.drop_vars(names=["hyai", "hybi", "hyam", "hybm"])

        # merge all datasets using the "ds_fl" dimensions as the join keys
        ds = xr.merge([ds_fl, ds_full, ds_surface, ds_rad], join="left")  # order matters!

        return ds

    def _calc_geopotential(self, ds: xr.Dataset) -> xr.DataArray:
        warnings.warn(
            (
                "The geopotential calculation implementation may assume the underlying grid "
                "starts at ground level. This may not be the case for IFS data. It may be "
                "better to use geometric height (altitude) instead of geopotential for downstream "
                "applications (tau cirrus, etc.)."
            ),
            UserWarning,
        )

        # TODO: this could be done via a mapping on the "lev" dimension
        # groupby("lev")

        z_level = ds["z"].copy()
        p_level = ds["surface_pressure"].copy()
        geopotential = xr.zeros_like(ds["t"])

        geopotential.attrs["standard_name"] = "geopotential"
        geopotential.attrs["units"] = "m**2 s**-2"
        geopotential.attrs["long_name"] = "Geopotential"

        # iterate through level layers from the bottom up
        for k in ds["lev"][::-1]:
            d_log_p = np.log(p_level / ds["air_pressure"].loc[dict(lev=k)])

            denom = p_level - ds["air_pressure"].loc[dict(lev=k)]
            alpha = 1 - d_log_p * ds["air_pressure"].loc[dict(lev=k)] / denom

            geopotential.loc[dict(lev=k)] = (
                z_level + ds["t_virtual"].loc[dict(lev=k)] * alpha * constants.R_d
            )

            # Update values for next loop
            z_level = geopotential.loc[dict(lev=k)].copy()
            p_level = ds["air_pressure"].loc[dict(lev=k)].copy()

        return geopotential



"""Cocip Grid Outputs."""

from __future__ import annotations

import numpy as np
import pandas as pd
import xarray as xr

from pycontrails.models.cocip import contrail_properties
from pycontrails.physics import constants, units

np.random.seed(1)


def cirrus_summary_statistics(
    df_contrails: pd.DataFrame,
    cloud_cover: xr.DataArray,
    tau_cirrus: xr.DataArray,
    tau_threshold: float = 0.1,
) -> pd.Series:
    """Calculate cirrus summary statistics.

    Calculates the (horizontal) gridded optical depth and fraction of cloud
    coverage arising from natural and
    contrail cirrus, as observed from above (i.e. satellites).


    Calculates the summary statistics, such as the percentage of:

    (1) total cirrus cover
    (2) natural cirrus cover
    (3) contrail cirrus cover; and
    (4) contrail cirrus cover under clear sky conditions for each time slice.

    The contrail cirrus cover, (3), is defined as the total cirrus cover
    (contrails + natural cirrus) minus the natural cirrus cover.

    Parameters
    ----------
    df_contrails : pd.DataFrame
        Aggregated contrail ouput.
        Note this must have a `flight_id` column with the `flight_id` of the causing flight.
    cloud_cover : xr.DataArray
        Description
    tau_cirrus : xr.DataArray
        Description
    tau_threshold : float, optional
        Description


    No Longer Returned
    ------------------
    pd.Series
        Description
    """
    cc_natural, tau_natural = _get_gridded_natural_cirrus_cover_and_tau(cloud_cover, tau_cirrus)
    tau_contrails_clear = get_gridded_tau_contrail(df_contrails, cc_natural)
    cc_contrails_clear = get_cirrus_coverage_from_tau(tau_contrails_clear, tau_threshold)

    tau_total = tau_contrails_clear + tau_natural
    cc_total = get_cirrus_coverage_from_tau(tau_total, tau_threshold)
    cc_contrails = cc_total - cc_natural

    # Get percentage cirrus cover
    lon_coords_met = tau_natural["longitude"].values
    lat_coords_met = tau_natural["latitude"].values
    pixel_area = _get_pixel_area(lon_coords_met, lat_coords_met)

    # This is specific to NATS
    # is_in_airspace = get_2d_airspace_domain_for_met_data(lon_coords_met, lat_coords_met)
    is_in_airspace = None

    cc_pct_total = _get_pct_cirrus_cover(cc_total, pixel_area, is_in_airspace)
    cc_pct_natural = _get_pct_cirrus_cover(cc_natural, pixel_area, is_in_airspace)
    cc_pct_contrails = _get_pct_cirrus_cover(cc_contrails, pixel_area, is_in_airspace)
    cc_pct_contrails_clear_sky = _get_pct_cirrus_cover(
        cc_contrails_clear, pixel_area, is_in_airspace
    )

    # Concatenate statistics
    df_cc_pct = {
        "Total Cirrus Cover Pct": cc_pct_total[0],
        "Natural Cirrus Cover Pct": cc_pct_natural[0],
        "Contrail Cirrus Cover Pct": cc_pct_contrails[0],
        "Contrail Cirrus Cover Pct - Clear Sky": cc_pct_contrails_clear_sky[0],
    }

    return pd.DataFrame(df_cc_pct)


def _get_gridded_natural_cirrus_cover_and_tau(
    cloud_cover: xr.DataArray, tau_cirrus: xr.DataArray
) -> tuple[xr.DataArray, xr.DataArray]:
    # Load DataArray to memory
    cc_3d = xr.DataArray.load(cloud_cover).copy()
    tau_cirrus_3d = xr.DataArray.load(tau_cirrus).copy()

    # Get gridded natural cirrus cover and optical depth (2D)
    cc_natural, tau_natural = get_max_cirrus_cover_and_tau(
        cc_3d, tau_cirrus_3d, boost_res=True, grid_res=0.05
    )

    return cc_natural, tau_natural


def get_cirrus_coverage_from_tau(tau_cirrus: xr.DataArray, tau_threshold: float) -> xr.DataArray:
    """Calculate the 2D cirrus coverage for the given `tau_cirrus` as observed by satellites.

    Note that the sensitivity of detecting cirrus is instrument dependent, and a tau_threshold
    is currently set at 0.1 to match the capability of satellites.
    """
    cirrus_cover = np.zeros_like(tau_cirrus)
    is_covered_by_cirrus = tau_cirrus.values > tau_threshold
    cirrus_cover[is_covered_by_cirrus] = 1
    da_cirrus_cover = xr.DataArray(data=cirrus_cover, coords=tau_cirrus.coords)
    return da_cirrus_cover


def _get_pct_cirrus_cover(
    cirrus_cover: xr.DataArray,
    pixel_area: xr.DataArray,
    is_in_airspace: xr.DataArray | None = None,
) -> pd.DataFrame:
    """Calculate the percentage cirrus coverage from the 2D cirrus coverage.

    Note that the global/regional cloud cover is calculated by summing
    the area of pixel cells that is covered by cirrus and dividing it
    by the total horizontal area.
    """
    if is_in_airspace is None:
        is_in_airspace = xr.ones_like(pixel_area, dtype=bool)

    cc_pct = np.array(
        [
            _get_pct_cirrus_cover_t(cirrus_cover.sel(time=t), pixel_area, is_in_airspace)
            for t in cirrus_cover["time"]
        ]
    )
    df_cc_pct = pd.DataFrame(cc_pct)
    df_cc_pct.index = cirrus_cover["time"].values
    return df_cc_pct


def _get_pct_cirrus_cover_t(
    cirrus_cover_2d: xr.DataArray, pixel_area: xr.DataArray, is_in_airspace: xr.DataArray
) -> float:
    """Calculate the percentage cirrus coverage for one time step.

    This is a helper function for `get_pct_cirrus_cover`.
    """
    cirrus_cover_2d_np = cirrus_cover_2d.values
    pixel_area_np = pixel_area.values
    is_in_airspace_np = is_in_airspace.values

    numer = np.sum(pixel_area_np[is_in_airspace_np] * cirrus_cover_2d_np[is_in_airspace_np])
    denom = np.sum(pixel_area_np[is_in_airspace_np])
    return numer / denom * 100


# ---
# helpers.py
# ---


def _get_pixel_area(lon_deg: np.ndarray, lat_deg: np.ndarray) -> xr.DataArray:
    """Calculate the area that is covered by the pixel of each grid cell.

    Source: https://www.pmel.noaa.gov/maillists/tmap/ferret_users/fu_2004/msg00023.html
    """
    lon_2d, lat_2d = np.meshgrid(lon_deg, lat_deg)
    d_lon = np.abs(lon_deg[1] - lon_deg[0])
    d_lat = np.abs(lat_deg[1] - lat_deg[0])

    area_lat_btm = _get_area_between_latitude_and_north_pole(lat_2d - d_lat)
    area_lat_top = _get_area_between_latitude_and_north_pole(lat_2d)
    grid_surface_area = (d_lon / 360) * (area_lat_btm - area_lat_top)
    return xr.DataArray(
        grid_surface_area.T,
        dims=["longitude", "latitude"],
        coords={"longitude": lon_deg, "latitude": lat_deg},
    )


def _get_area_between_latitude_and_north_pole(latitude: np.ndarray) -> np.ndarray:
    lat_radians = units.degrees_to_radians(latitude)
    return 2 * np.pi * constants.radius_earth**2 * (1 - np.sin(lat_radians))


# ---
# contrail_cirrus.py
# ---


def get_gridded_tau_contrail(
    df_contrails: pd.DataFrame,
    cc_natural_2d: xr.DataArray,
    *,
    time_slices: np.ndarray | None = None,
) -> xr.DataArray:
    """
    Convert the contrail waypoints to a gridded format.

    Notes
    -----
    The exact methodology of this function is outlined in Appendix A12 in Schumann (2012).

    References
    ----------
    Schumann, U., 2012. A contrail cirrus prediction model.
        Geoscientific Model Development, 5(3), pp.543-580.
    """
    # Get coordinates
    lon_coords_met = cc_natural_2d["longitude"].values
    lat_coords_met = cc_natural_2d["latitude"].values

    if time_slices is None:
        time_slices_met = cc_natural_2d["time"].values
    else:
        time_slices_met = np.copy(time_slices)

    # calculate edges for use in gridded calculations
    (
        df_contrails["lon_edge_l"],
        df_contrails["lat_edge_l"],
        df_contrails["lon_edge_r"],
        df_contrails["lat_edge_r"],
    ) = contrail_properties.contrail_edges(
        df_contrails["longitude"],
        df_contrails["latitude"],
        df_contrails["sin_a"],
        df_contrails["cos_a"],
        df_contrails["width"],
    )

    # Filter required contrail properties
    df_contrails = df_contrails[
        [
            "flight_id",  # TODO: do we really need this for the sort below?
            "waypoint",
            "continuous",
            "longitude",
            "latitude",
            "lon_edge_l",
            "lon_edge_r",
            "lat_edge_l",
            "lat_edge_r",
            "time",
            "width",
            "tau_contrail",
            "rf_net",
        ]
    ].copy()
    df_contrails["time"] = pd.to_datetime(df_contrails["time"])

    da_tau_contrail: list[xr.DataArray] = []
    for t in time_slices_met:
        da_tau_contrail_t = _get_gridded_outputs_time_slice(
            df_contrails, t, lon_coords_met, lat_coords_met
        )
        da_tau_contrail.append(da_tau_contrail_t)

    da_tau_contrail_xr = xr.concat(da_tau_contrail, dim="time")
    dim_order: list[str] = ["longitude", "latitude", "time"]
    return da_tau_contrail_xr.transpose(*dim_order)


def _get_gridded_outputs_time_slice(
    df_contrails: pd.DataFrame,
    time_slice: np.datetime64,
    lon_coords_met: np.ndarray,
    lat_coords_met: np.ndarray,
) -> xr.DataArray:
    """Convert the contrail waypoints for a given hour to a gridded format."""
    da_main_tau_contrail = _initialize_main_grid_xr(lon_coords_met, lat_coords_met)

    if len(df_contrails) == 0:
        return da_main_tau_contrail.assign_coords({"time": time_slice})

    contrail_heads_t = df_contrails[df_contrails["time"] == time_slice].copy()
    contrail_heads_t.sort_values(["flight_id", "waypoint"], inplace=True)
    contrail_tails_t = contrail_heads_t.shift(periods=-1)

    is_continuous = contrail_heads_t["continuous"].values
    contrail_heads_t = contrail_heads_t[is_continuous].copy()
    contrail_tails_t = contrail_tails_t[is_continuous].copy()

    for i in range(len(contrail_heads_t)):
        tau_contrail = get_contrail_segment_subgrid(
            contrail_heads_t.iloc[i], contrail_tails_t.iloc[i], da_main_tau_contrail
        )
        da_main_tau_contrail = _concatenate_segment_to_main_grid(da_main_tau_contrail, tau_contrail)

    return da_main_tau_contrail.expand_dims().assign_coords({"time": time_slice})


def _initialize_main_grid_xr(
    lon_coords_met: np.ndarray, lat_coords_met: np.ndarray
) -> xr.DataArray:
    """Initialize DataArray with specified coordinates.

    Initialize an empty xr.DataArray with longitude and latitude coordinates
    that is provided by the maximum cirrus coverage met variable. This
    "main grid" will be used to store the aggregated properties of the individual
    contrail segments, such as the gridded contrail optical depth and radiative forcing.
    """
    lon_grid, lat_grid = np.meshgrid(lon_coords_met, lat_coords_met)
    return xr.DataArray(
        np.zeros_like(lon_grid.T),
        dims=["longitude", "latitude"],
        coords={"longitude": lon_coords_met, "latitude": lat_coords_met},
    )


def _concatenate_segment_to_main_grid(
    da_vals_main_grid: xr.DataArray, da_vals_subgrid: xr.DataArray
) -> xr.DataArray:
    """Add values of the subgrid (contrail segment contribution to each pixel) to the main grid."""
    vals_main_grid = da_vals_main_grid.values
    vals_subgrid = da_vals_subgrid.values

    lon_main = da_vals_main_grid["longitude"].values
    lat_main = da_vals_main_grid["latitude"].values

    lon_subgrid = da_vals_subgrid["longitude"].values
    lat_subgrid = da_vals_subgrid["latitude"].values

    try:
        # NOTE: mypy is concerned that np.argmax may return something that is not be coercible
        # to a python int. So we cast explicitly here, and if something goes wrong, an error will
        # be raised
        ix_ = int(np.argmax(lon_main == lon_subgrid[0]))
        ix = int(np.argmax(lon_main == lon_subgrid[-1]) + 1)
        iy_ = int(np.argmax(lat_main == lat_subgrid[0]))
        iy = int(np.argmax(lat_main == lat_subgrid[-1]) + 1)
    except IndexError:
        # FIXME: log instead of print
        print(
            "NOTE: Contrail segment ignored as it is located beyond the meteorological boundaries. "
        )
    else:
        vals_main_grid[ix_:ix, iy_:iy] = vals_main_grid[ix_:ix, iy_:iy] + vals_subgrid

    return xr.DataArray(
        vals_main_grid,
        dims=["longitude", "latitude"],
        coords={"longitude": lon_main, "latitude": lat_main},
    )


# ---
# contrail_subgrid.py
# ---


def get_contrail_segment_subgrid(
    contrail_head: pd.Series, contrail_tail: pd.Series, da_main_grid: xr.DataArray
) -> xr.DataArray:
    """Convert the properties of an individual contrail segment to a grid format.

    Parameters
    ----------
    contrail_head : pd.Series
        Description
    contrail_tail : pd.Series
        Description
    da_main_grid : xr.DataArray
        Description
    """
    grid = _initialize_contrail_segment_subgrid(contrail_head, contrail_tail, da_main_grid)
    weights = _get_weights_to_subgrid_pixels(contrail_head, contrail_tail, grid)
    dist_perp = _get_perpendicular_dist_to_pixels(contrail_head, contrail_tail, weights)
    concentration = _get_gaussian_plume_concentration(
        contrail_head, contrail_tail, weights, dist_perp
    )
    tau_contrail = _get_subgrid_contrail_property(
        contrail_head, contrail_tail, weights, concentration
    )

    return tau_contrail


def _initialize_contrail_segment_subgrid(
    contrail_head: pd.Series, contrail_tail: pd.Series, da_main_grid: xr.DataArray
) -> xr.DataArray:
    """Initialize a contrail segment subgrid.

    Function initializes an empty xr.DataArray with a subset of
    longitude and latitude coordinates from the main grid. The spatial
    domain of the subgrid is defined by the area that is covered by the
    contrail. Note that the subgrid architecture is used to reduce the
    computational requirements.
    """
    # Longitude coordinates that are covered by the contrail segment
    lon_edges = np.array(
        [
            contrail_head["lon_edge_l"],
            contrail_head["lon_edge_r"],
            contrail_tail["lon_edge_l"],
            contrail_tail["lon_edge_r"],
        ]
    )
    lon_coords_main = da_main_grid["longitude"].values
    lon_tolerance = np.round(0.5 * (lon_coords_main[1] - lon_coords_main[0]), 3)
    is_near_contrail = (lon_coords_main >= (min(lon_edges) - lon_tolerance)) & (
        lon_coords_main <= (max(lon_edges) + lon_tolerance)
    )
    lon_coords_subgrid = lon_coords_main[is_near_contrail]

    # Latitude coordinates that are covered by the contrail segment
    lat_edges = np.array(
        [
            contrail_head["lat_edge_l"],
            contrail_head["lat_edge_r"],
            contrail_tail["lat_edge_l"],
            contrail_tail["lat_edge_r"],
        ]
    )
    lat_coords_main = da_main_grid["latitude"].values
    lat_tolerance = np.round(0.5 * (lat_coords_main[1] - lat_coords_main[0]), 3)
    is_near_contrail = (lat_coords_main >= (min(lat_edges) - lat_tolerance)) & (
        lat_coords_main <= (max(lat_edges) + lat_tolerance)
    )
    lat_coords_subgrid = lat_coords_main[is_near_contrail]

    # Initialize subgrid
    lon_subgrid, lat_subgrid = np.meshgrid(lon_coords_subgrid, lat_coords_subgrid)
    da_subgrid = xr.DataArray(
        np.zeros_like(lon_subgrid.T),
        dims=["longitude", "latitude"],
        coords={"longitude": lon_coords_subgrid, "latitude": lat_coords_subgrid},
    )
    return da_subgrid


def _get_weights_to_subgrid_pixels(
    contrail_head: pd.Series, contrail_tail: pd.Series, grid: xr.DataArray
) -> xr.DataArray:
    """Calculate weights for subgrid pixels.

    Function calculates the weights (from the beginning of the contrail segment)
    to the nearest longitude and latitude pixel in the subgrid. Note that weights
    with values that is out of range (w < 0 and w > 1) imply that the contrail
    segment do not contribute to the pixel. The methodology is in Appendix A12
    of Schumann (2012).
    """
    dx = units.longitude_distance_to_m(
        np.abs(contrail_head["longitude"] - contrail_tail["longitude"]),
        0.5 * (contrail_head["latitude"] + contrail_tail["latitude"]),
    )
    dy = units.latitude_distance_to_m(contrail_head["latitude"] - contrail_tail["latitude"])
    det = dx**2 + dy**2

    lon_subgrid, lat_subgrid = np.meshgrid(grid["longitude"].values, grid["latitude"].values)
    dx_grid = units.longitude_distance_to_m(
        np.abs(contrail_head["longitude"] - lon_subgrid),
        0.5 * (contrail_head["latitude"] + lat_subgrid),
    )
    dy_grid = units.latitude_distance_to_m(np.abs(contrail_head["latitude"] - lat_subgrid))
    weights = (dx * dx_grid + dy * dy_grid) / det
    return xr.DataArray(data=weights.T, coords=grid.coords)


def _get_perpendicular_dist_to_pixels(
    contrail_head: pd.Series, contrail_tail: pd.Series, weights: xr.DataArray
) -> xr.DataArray:
    """Calculate the perpendicular distance from the contrail segment to each pixel in the subgrid.

    Refer to Figure A7 in Schumann (2012).
    """
    lon_subgrid, lat_subgrid = np.meshgrid(weights["longitude"].values, weights["latitude"].values)

    # Longitude and latitude along contrail segment
    lon_s = contrail_head["longitude"] + weights.T.values * (
        contrail_tail["longitude"] - contrail_head["longitude"]
    )
    lat_s = contrail_head["latitude"] + weights.T.values * (
        contrail_tail["latitude"] - contrail_head["latitude"]
    )

    lon_dist = units.longitude_distance_to_m(
        np.abs(lon_s - lon_subgrid), 0.5 * (lat_s + lat_subgrid)
    )
    lat_dist = units.latitude_distance_to_m(np.abs(lat_s - lat_subgrid))
    dist_perp = (lon_dist**2 + lat_dist**2) ** 0.5
    da_dist_perp = xr.DataArray(dist_perp.T, coords=weights.coords)

    return da_dist_perp


def _get_gaussian_plume_concentration(
    contrail_head: pd.Series,
    contrail_tail: pd.Series,
    weights: xr.DataArray,
    dist_perp: xr.DataArray,
) -> xr.DataArray:
    """Calculate the relative concentration distribution along the axis of the contrail width.

    Assumes a one-dimensional Gaussian plume. The methodology for this function can be found in
    Appendix A11 of Schumann (2012).
    """

    width = weights.values * contrail_tail["width"] + (1 - weights.values) * contrail_head["width"]
    sigma_yy = 0.125 * width**2

    concentration = (4 / np.pi) ** 0.5 * np.exp(-0.5 * dist_perp.values**2 / sigma_yy)
    is_out_of_range = (weights.values < 0) | (weights.values > 1)
    concentration[is_out_of_range] = 0
    return xr.DataArray(concentration, coords=weights.coords)


def _get_subgrid_contrail_property(
    contrail_head: pd.Series,
    contrail_tail: pd.Series,
    weights: xr.DataArray,
    concentration: xr.DataArray,
) -> xr.DataArray:
    """Calculate the contrail segment contribution to each pixel in the subgrid."""
    w_1 = weights.values
    w_2 = 1 - w_1
    conc = concentration.values
    da_property_head = xr.ones_like(weights) * contrail_head["tau_contrail"]
    da_property_tail = xr.ones_like(weights) * contrail_tail["tau_contrail"]
    return (w_1 * da_property_tail + w_2 * da_property_head) * conc


# ---
# natural_cirrus.py
# ---


def get_max_cirrus_cover_and_tau(
    cloud_cover: xr.DataArray,
    tau_cirrus: xr.DataArray,
    *,
    boost_res: bool = False,
    grid_res: float = 0.05,
) -> tuple[xr.DataArray, xr.DataArray]:
    """Calculate the maximum natural cirrus coverage and optical depth.

    Returns these properties in the xr.DataArray format.
    """
    cc_max = _get_2d_cirrus_cover(cloud_cover)
    tau_cirrus_max = tau_cirrus.sel(
        level=tau_cirrus["level"][-1]
    )  # Get maximum values at lowest layer

    if boost_res:
        cc_max, tau_cirrus_max = _boost_cirrus_cover_and_tau_resolution(
            cc_max, tau_cirrus_max, grid_res=grid_res
        )

    return cc_max, tau_cirrus_max


def _get_2d_cirrus_cover(cloud_cover: xr.DataArray) -> xr.DataArray:
    """Calculate 2D cirrus cover effective for observers from above.

    cc_max(x,y,t) = max[cc(x,y,z,t)].
    """
    cc_max_t = [
        _get_2d_cirrus_cover_time_slice(cloud_cover, tt) for tt in range(len(cloud_cover["time"]))
    ]
    cc_max = xr.concat(cc_max_t, dim="time")
    dim_order: list = ["longitude", "latitude", "time"]
    return cc_max.transpose(*dim_order)


def _get_2d_cirrus_cover_time_slice(cloud_cover: xr.DataArray, time_int: int) -> xr.DataArray:
    cirrus_cover_t = cloud_cover.sel(time=cloud_cover["time"][time_int])
    return cirrus_cover_t.max(dim="level")


def _boost_cirrus_cover_and_tau_resolution(
    cc_max: xr.DataArray, tau_cirrus_max: xr.DataArray, *, grid_res: float = 0.05
) -> tuple[xr.DataArray, xr.DataArray]:
    """Increase resolution of cirrus cover and optical depth via approximation method by Schumann.

    This is necessary because the existing spatial resolution is too coarse to resolve the
    contributions from relatively narrow contrails.

    References
    ----------
    - :cite:`schumannContrailCirrusPrediction2012`
    """
    time = cc_max.time.values
    lon_hi_res, lat_hi_res = _boost_spatial_resolution(
        cc_max.longitude.values, cc_max.latitude.values, grid_res=grid_res
    )
    n_rep = int(len(lon_hi_res) / len(cc_max.longitude.values))
    res = [
        _boost_cc_and_tau_res_t(
            cc_max.sel(time=t).values, tau_cirrus_max.sel(time=t).values, n_rep=n_rep
        )
        for t in time
    ]

    # Convert results to xr.DataArray
    dim_order: list = ["longitude", "latitude", "time"]
    cc_max_hi_res = np.array([res[i][0] for i in range(len(res))])
    cc_max_hi_res_xr = xr.DataArray(
        cc_max_hi_res,
        dims=["time", "longitude", "latitude"],
        coords={"time": time, "longitude": lon_hi_res, "latitude": lat_hi_res},
    )
    cc_max_hi_res_xr = cc_max_hi_res_xr.transpose(*dim_order)

    tauc_max_hi_res = np.array([res[i][1] for i in range(len(res))])
    tauc_max_hi_res_xr = xr.DataArray(
        tauc_max_hi_res,
        dims=["time", "longitude", "latitude"],
        coords={"time": time, "longitude": lon_hi_res, "latitude": lat_hi_res},
    )
    tauc_max_hi_res_xr = tauc_max_hi_res_xr.transpose(*dim_order)
    return cc_max_hi_res_xr, tauc_max_hi_res_xr


def _boost_spatial_resolution(
    lon_deg: np.ndarray, lat_deg: np.ndarray, *, grid_res: float = 0.05
) -> tuple[np.ndarray, np.ndarray]:
    """Increase the longitude and latitude grid resolution."""
    d_lon = np.abs(lon_deg[1] - lon_deg[0])
    d_lat = np.abs(lat_deg[1] - lat_deg[0])
    is_whole_number = ((d_lon / grid_res) - int(d_lon / grid_res)) == 0

    if d_lon != d_lat:
        raise RuntimeWarning(
            "Note: Resolution of existing longitude and latitude inputs is not equal. "
        )

    if (d_lon <= grid_res) & (d_lat <= grid_res):
        raise ArithmeticError(
            "Error: Resolution of existing longitude or latitude is already higher than"
            ' "grid_res". '
        )

    if ~is_whole_number:
        raise ArithmeticError(
            "Select a grid resolution that provides a whole number to the new number of pixels. "
        )

    n_pixel_per_box = int(d_lon / grid_res)
    edges = grid_res * 0.5 * (n_pixel_per_box - 1)
    lon_boosted = np.linspace(
        min(lon_deg) - edges, max(lon_deg) + edges, (len(lon_deg) * n_pixel_per_box)
    )
    lat_boosted = np.linspace(
        min(lat_deg) - edges, max(lat_deg) + edges, (len(lat_deg) * n_pixel_per_box)
    )

    is_float_residual = np.abs(lon_boosted) < 1e-8
    lon_boosted[is_float_residual] = 0.0

    is_float_residual = np.abs(lat_boosted) < 1e-8
    lat_boosted[is_float_residual] = 0.0
    return lon_boosted, lat_boosted


def _boost_cc_and_tau_res_t(
    cc_max_t: np.ndarray, tau_cirrus_max_t: np.ndarray, *, n_rep: int, tau_threshold: float = 0.1
) -> tuple[np.ndarray, np.ndarray]:
    """Scale the mean value of the boosted resolution to the original pixel mean.

    The natural cirrus coverage and optical depth for the boosted resolution is
    distributed randomly in each pixel. This ensures that the mean value in the
    boosted resolution is equal to the value of the original pixel.

    References
    ----------
    - :cite:`schumannContrailCirrusPrediction2012`
    """
    cc_hi_res_t = _repeat_elements_along_rows_columns(cc_max_t, n_rep=n_rep)
    tau_cirrus_hi_res_t = _repeat_elements_along_rows_columns(tau_cirrus_max_t, n_rep=n_rep)

    cc_hi_res_adj_t = np.zeros_like(cc_hi_res_t)
    tauc_hi_res_adj_t = np.zeros_like(tau_cirrus_hi_res_t)

    # Adjusted cirrus optical depth
    rand_num = np.random.uniform(0, 1, np.shape(cc_hi_res_t))
    dx = 0.03  # Prevent division of small values: calibrated to match the original cirrus cover
    has_cover = rand_num > (1 + dx - cc_hi_res_t)
    tauc_hi_res_adj_t[has_cover] = tau_cirrus_hi_res_t[has_cover] / cc_hi_res_t[has_cover]

    # Adjusted cirrus coverage
    is_above_threshold = tauc_hi_res_adj_t >= tau_threshold
    cc_hi_res_adj_t[is_above_threshold] = 1
    return cc_hi_res_adj_t, tauc_hi_res_adj_t


def _repeat_elements_along_rows_columns(array_2d: np.ndarray, *, n_rep: int) -> np.ndarray:
    """Repeat the elements of a 2D numpy array n_rep times along each row and column."""
    dimension = np.shape(array_2d)

    # Repeating elements along axis=1
    array_2d_rep = [np.repeat(array_2d[i, :], n_rep) for i in np.arange(dimension[0])]
    stacked = np.vstack(array_2d_rep)

    # Repeating elements along axis=0
    return np.repeat(stacked, n_rep, axis=0)



"""ECMWF ERA5 data access."""

from __future__ import annotations

import hashlib
import logging
import os
import pathlib
import warnings
from contextlib import ExitStack
from datetime import datetime
from typing import TYPE_CHECKING, Any

LOG = logging.getLogger(__name__)

import pandas as pd
import xarray as xr
from overrides import overrides

import pycontrails
from pycontrails.core import cache, datalib
from pycontrails.core.met import MetDataset, MetVariable
from pycontrails.datalib.ecmwf.common import ECMWFAPI, rad_accumulated_to_average
from pycontrails.datalib.ecmwf.variables import (
    PRESSURE_LEVEL_VARIABLES,
    SURFACE_VARIABLES,
    TOAIncidentSolarRadiation,
    TopNetSolarRadiation,
    TopNetThermalRadiation,
)
from pycontrails.utils.temp import temp_file

if TYPE_CHECKING:
    import cdsapi


class ERA5(ECMWFAPI):
    """Class to support ERA5 data access, download, and organization.

    Requires account with
    `Copernicus Data Portal <https://cds.climate.copernicus.eu/cdsapp#!/home>`_
    and local credentials.

    API credentials can be stored in a ``~/.cdsapirc`` file
    or as ``CDSAPI_URL`` and ``CDSAPI_KEY`` environment variables.

        export CDSAPI_URL=...
        export CDSAPI_KEY=...

    Credentials can also be provided directly ``url`` and ``key`` keyword args.

    See `cdsapi <https://github.com/ecmwf/cdsapi>`_ documentation
    for more information.

    Parameters
    ----------
    time : datalib.TimeInput | None
        The time range for data retrieval, either a single datetime or (start, end) datetime range.
        Input must be datetime-like or tuple of datetime-like
        (`datetime`, :class:`pd.Timestamp`, :class:`np.datetime64`)
        specifying the (start, end) of the date range, inclusive.
        Datafiles will be downloaded from CDS for each day to reduce requests.
        If None, ``paths`` must be defined and all time coordinates will be loaded from files.
    variables : datalib.VariableInput
        Variable name (i.e. "t", "air_temperature", ["air_temperature, relative_humidity"])
    pressure_levels : datalib.PressureLevelInput, optional
        Pressure levels for data, in hPa (mbar)
        Set to -1 for to download surface level parameters.
        Defaults to -1.
    paths : str | list[str] | pathlib.Path | list[pathlib.Path] | None, optional
        Path to CDS NetCDF files to load manually.
        Can include glob patterns to load specific files.
        Defaults to None, which looks for files in the :attr:`cachestore` or CDS.
    timestep_freq : str, optional
        Manually set the timestep interval within the bounds defined by :attr:`time`.
        Supports any string that can be passed to `pd.date_range(freq=...)`.
        By default, this is set to "1H" for reanalysis products and "3H" for ensemble products.
    product_type : str, optional
        Product type, one of "reanalysis", "ensemble_mean", "ensemble_members", "ensemble_spread"
    grid : float, optional
        Specify latitude/longitude grid spacing in data.
        By default, this is set to 0.25 for reanalysis products and 0.5 for ensemble products.
    cachestore : cache.CacheStore | None, optional
        Cache data store for staging ECMWF ERA5 files.
        Defaults to :class:`cache.DiskCacheStore`.
        If None, cache is turned off.
    url : str
        Override `cdsapi <https://github.com/ecmwf/cdsapi>`_ url
    key : str
        Override `cdsapi <https://github.com/ecmwf/cdsapi>`_ key

    Notes
    -----
    ERA5 parameter list:
    https://confluence.ecmwf.int/pages/viewpage.action?pageId=82870405#ERA5:datadocumentation-Parameterlistings

    All accumulated radiative quantities are converted to average instantaneous quantities.
    See https://www.ecmwf.int/sites/default/files/elibrary/2015/18490-radiation-quantities-ecmwf-model-and-mars.pdf
    for more information.

    Local ``paths`` are loaded using :func:`xarray.open_mfdataset`.
    Pass ``xr_kwargs`` inputs to :meth:`open_metdataset` to customize file loading.

    Examples
    --------
    >>> from datetime import datetime
    >>> from pycontrails.datalib.ecmwf import ERA5
    >>> from pycontrails import GCPCacheStore

    >>> # Store data files from CDS to local disk (default behavior)
    >>> era5 = ERA5(
    ...     "2020-06-01 12:00:00",
    ...     variables=["air_temperature", "relative_humidity"],
    ...     pressure_levels=[350, 300]
    ... )

    >>> # cache files to google cloud storage
    >>> gcp_cache = cache.GCPCacheStore(bucket="contrails-301217-unit-test", cache_dir="ecmwf")
    >>> era5 = ERA5(
    ...     "2020-06-01 12:00:00",
    ...     variables=["air_temperature", "relative_humidity"],
    ...     pressure_levels=[350, 300],
    ...     cachestore=gcp_cache
    ... )
    """  # noqa: E501

    __slots__ = (
        "product_type",
        "cds",
        "url",
        "key",
    )

    #: Product type, one of "reanalysis", "ensemble_mean", "ensemble_members", "ensemble_spread"
    product_type: str

    #: Handle to ``cdsapi.Client``
    cds: cdsapi.Client

    #: User provided ``cdsapi.Client`` url
    url: str | None

    #: User provided ``cdsapi.Client`` url
    key: str | None

    __marker = object()

    def __init__(
        self,
        time: datalib.TimeInput | None,
        variables: datalib.VariableInput,
        pressure_levels: datalib.PressureLevelInput = -1,
        paths: str | list[str] | pathlib.Path | list[pathlib.Path] | None = None,
        timestep_freq: str = "1H",
        product_type: str = "reanalysis",
        grid: float = 0.25,
        cachestore: cache.CacheStore | None = __marker,  # type: ignore[assignment]
        url: str | None = None,
        key: str | None = None,
    ):
        # inputs
        self.product_type = product_type
        if cachestore is self.__marker:
            cachestore = cache.DiskCacheStore()
        self.cachestore = cachestore
        self.paths = paths
        self.url = url or os.environ.get("CDSAPI_URL", None)
        self.key = key or os.environ.get("CDSAPI_KEY", None)

        if time is None and paths is None:
            raise ValueError("Time input is required when paths is None")

        product_types = ("reanalysis", "ensemble_mean", "ensemble_members", "ensemble_spread")
        if product_type not in product_types:
            raise ValueError(
                f"Unknown product_type {product_type}. "
                f"Currently support product types: {', '.join(product_types)}"
            )

        if product_type == "reanalysis":
            grid_min = 0.25
        else:
            grid_min = 0.5

        msg = (
            f"The smallest resolution available through the CDS API is {grid_min} degrees. "
            f"Your downloaded data will have resolution {grid}, but it is a reinterpolation "
            f"of the {grid_min} degree data. The same interpolation can be achieved "
            "directly with `xarray`."
        )

        # process inputs
        if "ensemble" in product_type:
            # If using a nonstandard product and default timestep_freq, override default
            if timestep_freq == "1H":
                timestep_freq = "3H"

            # And taking the same procedure for grid
            if grid == 0.25:
                grid = 0.5
            elif grid < 0.5:
                warnings.warn(msg)

        elif grid < 0.25:
            warnings.warn(msg)

        self.grid = grid
        self.timesteps = datalib.parse_timesteps(time, freq=timestep_freq)
        self.pressure_levels = datalib.parse_pressure_levels(
            pressure_levels, self.supported_pressure_levels
        )
        self.variables = datalib.parse_variables(variables, self.supported_variables)

        # ensemble_mean, etc - time is only available on the 0, 3, 6, etc
        if "ensemble" in product_type:
            if {t.hour for t in self.timesteps} - {0, 3, 6, 9, 12, 15, 18, 21}:
                raise NotImplementedError("Ensemble products only support every three hours")

    def __repr__(self) -> str:
        base = super().__repr__()
        return f"{base}\n\tDataset: {self.dataset}\n\tProduct type: {self.product_type}"

    @property
    def hash(self) -> str:
        """Generate a unique hash for this datasource.

        Returns
        -------
        str
            Unique hash for met instance (sha1)
        """
        hashstr = (
            f"{self.__class__.__name__}{self.timesteps}{self.variable_shortnames}"
            f"{self.pressure_levels}{self.grid}{self.product_type}"
        )
        return hashlib.sha1(bytes(hashstr, "utf-8")).hexdigest()

    @property
    def pressure_level_variables(self) -> list[MetVariable]:
        """ECMWF pressure level parameters.

        Returns
        -------
        list[MetVariable] | None
            List of MetVariable available in datasource
        """
        return PRESSURE_LEVEL_VARIABLES

    @property
    def single_level_variables(self) -> list[MetVariable]:
        """ECMWF surface level parameters.

        Returns
        -------
        list[MetVariable] | None
            List of MetVariable available in datasource
        """
        return SURFACE_VARIABLES

    @property
    def supported_pressure_levels(self) -> list[int]:
        """Get pressure levels available from ERA5 pressure level dataset.

        Returns
        -------
        list[int]
            List of integer pressure level values
        """
        return [
            1000,
            975,
            950,
            925,
            900,
            875,
            850,
            825,
            800,
            775,
            750,
            700,
            650,
            600,
            550,
            500,
            450,
            400,
            350,
            300,
            250,
            225,
            200,
            175,
            150,
            125,
            100,
            70,
            50,
            30,
            20,
            10,
            7,
            5,
            3,
            2,
            1,
            -1,
        ]

    @property
    def dataset(self) -> str:
        """Select dataset for download based on :attr:`pressure_levels`.

        One of "reanalysis-era5-pressure-levels" or "reanalysis-era5-single-levels"

        Returns
        -------
        str
            ERA5 dataset name in CDS
        """
        if self.pressure_levels != [-1]:
            return "reanalysis-era5-pressure-levels"
        return "reanalysis-era5-single-levels"

    def create_cachepath(self, t: datetime | pd.Timestamp) -> str:
        """Return cachepath to local ERA5 data file based on datetime.

        This uniquely defines a cached data file ith class parameters.

        Parameters
        ----------
        t : datetime | pd.Timestamp
            Datetime of datafile

        Returns
        -------
        str
            Path to local ERA5 data file
        """
        if self.cachestore is None:
            raise ValueError("self.cachestore attribute must be defined to create cache path")

        datestr = t.strftime("%Y%m%d-%H")

        # set date/time for file
        if self.pressure_levels == [-1]:
            suffix = f"era5sl{self.grid}{self.product_type}"
        else:
            suffix = f"era5pl{self.grid}{self.product_type}"

        # return cache path
        return self.cachestore.path(f"{datestr}-{suffix}.nc")

    @overrides
    def download_dataset(self, times: list[datetime]) -> None:
        download_times: dict[datetime, list[datetime]] = {}
        for t in times:
            unique_day = datetime(t.year, t.month, t.day)
            download_times.setdefault(unique_day, [])

            download_times[unique_day].append(t)

        # download data file for each unique day
        LOG.debug(f"Downloading ERA5 dataset for times {times}")
        for unique_day in download_times:
            self._download_file(download_times[unique_day])

    @overrides
    def cache_dataset(self, dataset: xr.Dataset) -> None:
        if self.cachestore is None:
            LOG.debug("Cache is turned off, skipping")
            return

        with ExitStack() as stack:
            # group by hour and save one dataset for each hour to temp file
            time_group, datasets = zip(*dataset.groupby("time", squeeze=False))

            xarray_temp_filenames = [stack.enter_context(temp_file()) for _ in time_group]
            xr.save_mfdataset(datasets, xarray_temp_filenames)

            # put each hourly file into cache
            cache_path = [
                self.create_cachepath(datetime.utcfromtimestamp(tg.tolist() / 1e9))
                for tg in time_group
            ]
            self.cachestore.put_multiple(xarray_temp_filenames, cache_path)

    @overrides
    def open_metdataset(
        self,
        dataset: xr.Dataset | None = None,
        xr_kwargs: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> MetDataset:
        xr_kwargs = xr_kwargs or {}

        #  short-circuit dataset or file paths if provided
        if dataset is not None:
            ds = self._preprocess_era5_dataset(dataset)

        # load from local paths
        elif self.paths is not None:
            ds = self._open_and_cache(xr_kwargs)

        # load from cache or download
        else:
            if self.cachestore is None:
                raise ValueError("Cachestore is required to download data")

            # confirm files are downloaded from CDS or MARS
            self.download(**xr_kwargs)

            # ensure all files are guaranteed to be available locally here
            # this would download a file from a remote (e.g. GCP) cache
            disk_cachepaths = [self.cachestore.get(f) for f in self._cachepaths]

            # run MetDataset constructor
            ds = self.open_dataset(disk_cachepaths, **xr_kwargs)

            # TODO: corner case
            # If any files are already cached, they will not have the version attached
            ds.attrs.setdefault("pycontrails_version", pycontrails.__version__)

        # run the same ECMWF-specific processing on the dataset
        mds = self._process_dataset(ds, **kwargs)

        # convert accumulated radiation values to average instantaneous values
        # set minimum for all values to 0

        # accumulations are 3 hours for ensembles, 1 hour for reanalysis
        if "ensemble" in self.product_type:
            dt_accumulation = 3 * 60 * 60
        else:
            dt_accumulation = 60 * 60

        for key in (
            TOAIncidentSolarRadiation.standard_name,
            TopNetSolarRadiation.standard_name,
            TopNetThermalRadiation.standard_name,
        ):
            if key in mds.data:
                rad_accumulated_to_average(mds, key, dt_accumulation)

        return mds

    def _open_and_cache(self, xr_kwargs: dict[str, Any]) -> xr.Dataset:
        """Open and cache :class:`xr.Dataset` from :attr:`self.paths`.

        Parameters
        ----------
        xr_kwargs : dict[str, Any]
            Additional kwargs passed directly to :func:`xarray.open_mfdataset`.
            See :meth:`open_metdataset`.

        Returns
        -------
        xr.Dataset
            Dataset opened from local paths.
        """

        if self.paths is None:
            raise ValueError("Attribute `self.paths` must be defined to open and cache")

        # if timesteps are defined and all timesteps are cached already
        # then we can skip loading
        if self.timesteps and self.cachestore and not self.list_timesteps_not_cached(**xr_kwargs):
            LOG.debug("All timesteps already in cache store")
            disk_cachepaths = [self.cachestore.get(f) for f in self._cachepaths]
            return self.open_dataset(disk_cachepaths, **xr_kwargs)

        ds = self.open_dataset(self.paths, **xr_kwargs)
        ds = self._preprocess_era5_dataset(ds)
        self.cache_dataset(ds)

        return ds

    def _download_file(self, times: list[datetime]) -> None:
        """Download data file for specific sets of times for *unique date* from CDS API.

        Splits datafiles by the hour and saves each hour in the cache datastore.
        Overwrites files if they already exists.

        Parameters
        ----------
        times : list[datetime]
            Times to download from single day
        """

        # set date/time for file
        date_str = times[0].strftime("%Y-%m-%d")

        # check to make sure times are all on the same day
        if any(dt.strftime("%Y-%m-%d") != date_str for dt in times):
            raise ValueError("All times must be on the same date when downloading from CDS")

        time_strs = [t.strftime("%H:%M") for t in times]

        # make request of cdsapi
        request: dict[str, Any] = {
            "product_type": self.product_type,
            "variable": self.variable_shortnames,
            "date": date_str,
            "time": time_strs,
            "grid": [self.grid, self.grid],
            "format": "netcdf",
        }
        if self.dataset == "reanalysis-era5-pressure-levels":
            request["pressure_level"] = self.pressure_levels

        # Open ExitStack to control temp_file context manager
        with ExitStack() as stack:
            # hold downloaded file in named temp file
            cds_temp_filename = stack.enter_context(temp_file())

            LOG.debug(f"Performing CDS request: {request} to dataset {self.dataset}")
            if not hasattr(self, "cds"):
                try:
                    import cdsapi
                except ModuleNotFoundError as e:
                    raise ModuleNotFoundError(
                        "Some `ecmwf` module dependencies are missing. "
                        "Please install all required dependencies using `pip install -e .[ecmwf]`"
                    ) from e

                try:
                    self.cds = cdsapi.Client(url=self.url, key=self.key)
                # cdsapi throws base-level Exception
                except Exception as err:
                    raise CDSCredentialsNotFound from err

            self.cds.retrieve(self.dataset, request, cds_temp_filename)

            # open file, edit, and save for each hourly time step
            ds = stack.enter_context(
                xr.open_dataset(cds_temp_filename, engine=datalib.NETCDF_ENGINE)
            )

            # run preprocessing before cache
            ds = self._preprocess_era5_dataset(ds)

            self.cache_dataset(ds)

    def _preprocess_era5_dataset(self, ds: xr.Dataset) -> xr.Dataset:
        """Process ERA5 data before caching.

        Parameters
        ----------
        ds : xr.Dataset
            Loaded :class:`xr.Dataset`

        Returns
        -------
        xr.Dataset
            Processed :class:`xr.Dataset`
        """

        if "pycontrails_version" in ds.attrs:
            LOG.debug("Input dataset processed with pycontrails > 0.29")
            return ds

        # not pre-processed source file from `download` or `paths`

        # for "reanalysis-era5-single-levels" or if self.pressure_levels length == 1,
        # then the netcdf file does not contain the dimension "level"
        if len(self.pressure_levels) == 1:
            ds = ds.expand_dims({"level": self.pressure_levels})

        ds.attrs["pycontrails_version"] = pycontrails.__version__
        return ds


class CDSCredentialsNotFound(Exception):
    """Raise when CDS credentials are not found by `cdsapi.Client` instance."""
